# ruff: noqa: F821
# mypy: ignore-errors
import functools as ft
import typing as t

import pandas as pd

try:
    import pandera as pda
    import pandera.typing as pt
except ModuleNotFoundError:
    from ... import utils

    utils.databricks.mock_pandera()

    import pandera as pda
    import pandera.typing as pt

# TODO: re-use existing fields from raw data schemas?

TermField = ft.partial(
    pda.Field,
    dtype_kwargs={
        "categories": ["FALL", "WINTER", "SPRING", "SUMMER"],
        # TODO: do these need to be ordered anymore?
        # if so, how should we handle schools with alt orderings?
        # "ordered": True,
    },
)
PlacementField = ft.partial(
    pda.Field, nullable=True, dtype_kwargs={"categories": ["C", "N"]}
)
GPAField = ft.partial(pda.Field, nullable=True, ge=0.0, le=4.0)
NumCreditsGt0Field = ft.partial(pda.Field, nullable=True, ge=0.0)


class PDPStudentTermsDataSchema(pda.DataFrameModel):
    student_id: pt.Series["string"]
    term_id: pt.Series["string"]
    institution_id: pt.Series["string"]
    academic_year: pt.Series["string"]
    academic_term: pt.Series[pd.CategoricalDtype] = TermField()
    term_start_dt: pt.Series["datetime64[ns]"]
    term_rank: pt.Series["int8"]
    term_in_peak_covid: pt.Series["bool"]
    term_rank_core: pt.Series["Int8"] = pda.Field(nullable=True)
    term_is_core: pt.Series["bool"]
    num_courses: pt.Series["int8"]
    num_courses_passed: pt.Series["int8"]
    num_courses_completed: pt.Series["int8"]
    num_credits_attempted: pt.Series["float32"]
    num_credits_earned: pt.Series["float32"]
    course_ids: pt.Series["object"]
    course_subjects: pt.Series["object"]
    course_subject_areas: pt.Series["object"]
    course_id_nunique: pt.Series["Int8"] = pda.Field(nullable=True)
    course_subject_nunique: pt.Series["Int8"] = pda.Field(nullable=True)
    course_subject_area_nunique: pt.Series["Int8"] = pda.Field(nullable=True)
    course_level_mean: pt.Series["Float32"] = pda.Field(nullable=True)
    course_level_std: pt.Series["Float32"] = pda.Field(nullable=True)
    course_grade_numeric_mean: pt.Series["Float32"] = pda.Field(nullable=True)
    course_grade_numeric_std: pt.Series["Float32"] = pda.Field(nullable=True)
    section_num_students_enrolled_mean: pt.Series["Float32"] = pda.Field(nullable=True)
    section_num_students_enrolled_std: pt.Series["Float32"] = pda.Field(nullable=True)
    sections_num_students_enrolled: pt.Series["int8"]
    sections_num_students_passed: pt.Series["int8"]
    sections_num_students_completed: pt.Series["int8"]
    num_courses_enrolled_at_other_institution_s_y: pt.Series["int8"]
    num_courses_core_course_y: t.Optional[pt.Series["int8"]]
    num_courses_core_course_n: t.Optional[pt.Series["int8"]]
    num_courses_course_type_cc_cd: pt.Series["int8"]
    num_courses_course_type_cu: pt.Series["int8"]
    num_courses_course_type_cg: pt.Series["int8"]
    num_courses_course_type_cc: pt.Series["int8"]
    num_courses_course_type_cd: pt.Series["int8"]
    num_courses_course_type_el: pt.Series["int8"]
    num_courses_course_type_ab: pt.Series["int8"]
    num_courses_course_type_ge: pt.Series["int8"]
    num_courses_course_type_nc: pt.Series["int8"]
    num_courses_course_type_o: pt.Series["int8"]
    num_courses_delivery_method_f: pt.Series["int8"]
    num_courses_delivery_method_o: pt.Series["int8"]
    num_courses_delivery_method_h: pt.Series["int8"]
    num_courses_math_or_english_gateway_m: pt.Series["int8"]
    num_courses_math_or_english_gateway_e: pt.Series["int8"]
    num_courses_math_or_english_gateway_na: pt.Series["int8"]
    num_courses_co_requisite_course_y: pt.Series["int8"]
    num_courses_co_requisite_course_n: pt.Series["int8"]
    num_courses_course_instructor_employment_status_pt: pt.Series["int8"]
    num_courses_course_instructor_employment_status_ft: pt.Series["int8"]
    num_courses_course_instructor_rank_1: pt.Series["int8"]
    num_courses_course_instructor_rank_2: pt.Series["int8"]
    num_courses_course_instructor_rank_3: pt.Series["int8"]
    num_courses_course_instructor_rank_4: pt.Series["int8"]
    num_courses_course_instructor_rank_5: pt.Series["int8"]
    num_courses_course_instructor_rank_6: pt.Series["int8"]
    num_courses_course_instructor_rank_7: pt.Series["int8"]
    # NOTE: let's cover most likely values, not all possible values
    num_courses_course_level_1: pt.Series["int8"]
    num_courses_course_grade_a: pt.Series["int8"]
    num_courses_grade_is_failing_or_withdrawal: pt.Series["int8"]
    num_courses_grade_above_section_avg: pt.Series["int8"]
    cohort: pt.Series["string"]
    cohort_term: pt.Series[pd.CategoricalDtype] = TermField()
    # TODO: re-use existing fields from raw data schemas?
    enrollment_type: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True,
        dtype_kwargs={"categories": ["FIRST-TIME", "RE-ADMIT", "TRANSFER-IN"]},
    )
    enrollment_intensity_first_term: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True,
        dtype_kwargs={"categories": ["FULL-TIME", "PART-TIME"]},
    )
    math_placement: pt.Series[pd.CategoricalDtype] = PlacementField()
    english_placement: pt.Series[pd.CategoricalDtype] = PlacementField()
    reading_placement: pt.Series[pd.CategoricalDtype] = PlacementField()
    dual_and_summer_enrollment: pt.Series[pd.CategoricalDtype] = pda.Field(
        nullable=True,
        dtype_kwargs={"categories": ["DE", "SE", "DS"]},
    )
    # NOTE: intentionally not bothering with categoricals for student demographics
    student_age: pt.Series["string"]
    race: pt.Series["string"]
    ethnicity: pt.Series["string"]
    first_gen: pt.Series["string"]
    pell_status_first_year: pt.Series["string"] = pda.Field(nullable=True)
    special_program: t.Optional[pt.Series["string"]] = pda.Field(nullable=True)
    naspa_first_generation: t.Optional[pt.Series["string"]] = pda.Field(nullable=True)
    incarcerated_status: t.Optional[pt.Series["string"]] = pda.Field(nullable=True)
    military_status: t.Optional[pt.Series["string"]] = pda.Field(nullable=True)
    employment_status: t.Optional[pt.Series["string"]] = pda.Field(nullable=True)
    disability_status: t.Optional[pt.Series["string"]] = pda.Field(nullable=True)
    foreign_language_completion: t.Optional[pt.Series["string"]] = pda.Field(
        nullable=True
    )
    credential_type_sought_year_1: pt.Series["string"] = pda.Field(nullable=True)
    program_of_study_term_1: pt.Series["string"] = pda.Field(nullable=True)
    program_of_study_year_1: pt.Series["string"] = pda.Field(nullable=True)
    gpa_group_term_1: pt.Series["Float32"] = GPAField()
    gpa_group_year_1: pt.Series["Float32"] = GPAField()
    number_of_credits_earned_year_1: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_2: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_2: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_3: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_3: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_attempted_year_4: pt.Series["Float32"] = NumCreditsGt0Field()
    number_of_credits_earned_year_4: pt.Series["Float32"] = NumCreditsGt0Field()
    # NOTE: let's skip the many "outcome"-like columns here; we'll drop them later, anyway
    cohort_id: pt.Series["string"]
    cohort_start_dt: pt.Series["datetime64[ns]"]
    student_program_of_study_area_term_1: pt.Series["string"] = pda.Field(nullable=True)
    student_program_of_study_area_year_1: pt.Series["string"] = pda.Field(nullable=True)
    student_program_of_study_area_changed_term_1_to_year_1: pt.Series["boolean"] = (
        pda.Field(nullable=True)
    )
    diff_gpa_term_1_to_year_1: pt.Series["Float32"] = pda.Field(nullable=True)
    frac_credits_earned_year_1: pt.Series["Float32"] = pda.Field(nullable=True)
    frac_credits_earned_year_2: pt.Series["Float32"] = pda.Field(nullable=True)
    frac_credits_earned_year_3: pt.Series["Float32"] = pda.Field(nullable=True)
    frac_credits_earned_year_4: pt.Series["Float32"] = pda.Field(nullable=True)
    year_of_enrollment_at_cohort_inst: pt.Series["int8"]
    term_is_while_student_enrolled_at_other_inst: pt.Series["bool"]
    frac_credits_earned: pt.Series["Float32"] = pda.Field(nullable=True)
    student_term_enrollment_intensity: pt.Series["string"]
    frac_courses_passed: pt.Series["float32"]
    frac_courses_completed: pt.Series["float32"]
    frac_courses_enrolled_at_other_institution_s_y: pt.Series["float32"]
    frac_courses_core_course_y: t.Optional[pt.Series["float32"]]
    frac_courses_core_course_n: t.Optional[pt.Series["float32"]]
    frac_courses_course_type_cc_cd: pt.Series["float32"]
    frac_courses_course_type_cu: pt.Series["float32"]
    frac_courses_course_type_cg: pt.Series["float32"]
    frac_courses_course_type_cc: pt.Series["float32"]
    frac_courses_course_type_cd: pt.Series["float32"]
    frac_courses_course_type_el: pt.Series["float32"]
    frac_courses_course_type_ab: pt.Series["float32"]
    frac_courses_course_type_ge: pt.Series["float32"]
    frac_courses_course_type_nc: pt.Series["float32"]
    frac_courses_course_type_o: pt.Series["float32"]
    frac_courses_delivery_method_f: pt.Series["float32"]
    frac_courses_delivery_method_o: pt.Series["float32"]
    frac_courses_delivery_method_h: pt.Series["float32"]
    frac_courses_math_or_english_gateway_m: pt.Series["float32"]
    frac_courses_math_or_english_gateway_e: pt.Series["float32"]
    frac_courses_math_or_english_gateway_na: pt.Series["float32"]
    frac_courses_co_requisite_course_y: pt.Series["float32"]
    frac_courses_co_requisite_course_n: pt.Series["float32"]
    frac_courses_course_instructor_employment_status_pt: pt.Series["float32"]
    frac_courses_course_instructor_employment_status_ft: pt.Series["float32"]
    frac_courses_course_instructor_rank_1: pt.Series["float32"]
    frac_courses_course_instructor_rank_2: pt.Series["float32"]
    frac_courses_course_instructor_rank_3: pt.Series["float32"]
    frac_courses_course_instructor_rank_4: pt.Series["float32"]
    frac_courses_course_instructor_rank_5: pt.Series["float32"]
    frac_courses_course_instructor_rank_6: pt.Series["float32"]
    frac_courses_course_instructor_rank_7: pt.Series["float32"]
    # NOTE: let's cover most likely values, not all possible values
    frac_courses_course_level_1: pt.Series["float32"]
    frac_courses_course_grade_a: pt.Series["float32"]
    frac_courses_grade_is_failing_or_withdrawal: pt.Series["float32"]
    frac_courses_grade_above_section_avg: pt.Series["float32"]
    frac_sections_students_passed: pt.Series["float32"]
    frac_sections_students_completed: pt.Series["float32"]
    student_pass_rate_above_sections_avg: pt.Series["bool"]
    student_completion_rate_above_sections_avg: pt.Series["bool"]
    # NOTE: let's not bother with the many cumulative features (for now?)
    # cumnum_terms_enrolled: pt.Series["int8"]
    # etc.
    num_courses_diff_prev_term: pt.Series["Int8"] = pda.Field(nullable=True)
    num_credits_earned_diff_prev_term: pt.Series["Float32"] = pda.Field(nullable=True)
    course_grade_numeric_mean_diff_prev_term: pt.Series["Float32"] = pda.Field(
        nullable=True
    )
    num_courses_diff_term_1_to_term_2: pt.Series["Int8"] = pda.Field(nullable=True)
    num_credits_earned_diff_term_1_to_term_2: pt.Series["Float32"] = pda.Field(
        nullable=True
    )
    course_grade_numeric_mean_diff_term_1_to_term_2: pt.Series["Float32"] = pda.Field(
        nullable=True
    )

    class Config:
        coerce = True
        unique_column_names = True
        add_missing_columns = False
        drop_invalid_rows = False
        unique = ["student_id", "term_id"]



import numpy as np
import pandas as pd
import pytest

from student_success_tool.modeling import feature_selection as fs


@pytest.fixture(scope="module")
def df():
    return pd.DataFrame(
        {
            "id": ["01", "02", "03", "04", "05"],
            "categorical_feature": ["a", "a", "b", "c", "b"],
            "include_feature": [3.142, 1.414, 1.618, 2.718, 0.693],
            "good_feature": [1, 2, 3, 4, 5],
            "partially_incomplete_feature": [np.nan, np.nan, 8, 5, 3],
            "incomplete_feature": [np.nan, 0.0, np.nan, np.nan, np.nan],
            "low_variance_feature": [1, 1, 1, 1, 2],
            "zero_variance_feature": [1, 1, 1, 1, 1],
            "high_collinear_feature": [6, 7, 8, 9, 10],
            "low_collinear_feature": [6.5, 1.5, 4, 3.5, 2],
        }
    ).astype({"id": "string", "categorical_feature": "category"})


# TODO: figure out why collinaer features col keeps raising errors / gives bad results
# @pytest.mark.parametrize(
#     ["non_feature_cols", "force_include_cols", "exp_dropped_cols"],
#     [
#         (
#             ["id"],
#             ["include_feature"],
#             ["incomplete_feature", "zero_variance_feature", "high_collinear_feature"],
#         ),
#     ],
# )
# def test_select_features(df, non_feature_cols, force_include_cols, exp_dropped_cols):
#     df_result = fs.select_features(
#         df, non_feature_cols=non_feature_cols, force_include_cols=force_include_cols
#     )
#     assert df.columns.difference(df_result.columns).tolist() == exp_dropped_cols
#     assert df_result.shape[0] == df.shape[0]


@pytest.mark.parametrize(
    ["threshold", "exp_dropped_cols"],
    [
        (0.5, ["incomplete_feature"]),
        (0.4, ["incomplete_feature", "partially_incomplete_feature"]),
    ],
)
def test_drop_incomplete_features(df, threshold, exp_dropped_cols):
    df_result = fs.drop_incomplete_features(df, threshold=threshold)
    assert df.columns.difference(df_result.columns).tolist() == exp_dropped_cols
    assert df_result.shape[0] == df.shape[0]


@pytest.mark.parametrize(
    ["threshold", "exp_dropped_cols"],
    [
        (0.0, ["incomplete_feature", "zero_variance_feature"]),
        (0.2, ["incomplete_feature", "low_variance_feature", "zero_variance_feature"]),
    ],
)
def test_drop_low_variance_features(df, threshold, exp_dropped_cols):
    df_result = fs.drop_low_variance_features(df, threshold=threshold)
    assert df.columns.difference(df_result.columns).tolist() == exp_dropped_cols
    assert df_result.shape[0] == df.shape[0]


def test_drop_collinear_features_iteratively():
    related_features = pd.DataFrame(
        {"A": [1, 2, 3, 4, 5], "B": [6, 7, 8, 9, 10], "C": [6.5, 1.5, 4, 3.5, 2]}
    )
    returned_df = fs.drop_collinear_features_iteratively(
        related_features, force_include_cols=[]
    )
    assert returned_df.columns.tolist() == ["A", "C"]
    assert returned_df.shape[0] == related_features.shape[0]


def test_booleans_drop_collinear_features_iteratively():
    related_features = pd.DataFrame(
        {
            "A": [2, 12, 6.5, 4, 3.5],
            "bool": [True, False, True, False, True],
            "numeric": [1, 0, 1, 0, 1],
        }
    )
    returned_df = fs.drop_collinear_features_iteratively(
        related_features, force_include_cols=[]
    )
    assert returned_df.columns.tolist() == ["A", "numeric"]
    assert returned_df.shape[0] == related_features.shape[0]


def test_drop_collinear_features_iteratively_force_include():
    related_features = pd.DataFrame(
        {"A": [1, 2, 3, 4, 5], "B": [6, 7, 8, 9, 10], "C": [6.5, 1.5, 4, 3.5, 2]}
    )
    returned_df = fs.drop_collinear_features_iteratively(
        related_features, force_include_cols=["B"]
    )
    assert returned_df.columns.tolist() == ["B", "C"]
    assert returned_df.shape[0] == related_features.shape[0]


def test_drop_collinear_features_iteratively_fails_with_one_feature_left():
    related_features = pd.DataFrame({"A": [1, 2, 3, 4, 5], "B": [6, 7, 8, 9, 10]})
    with pytest.raises(Exception):
        fs.drop_collinear_features_iteratively(related_features)



import sys
import os
import json
from typing import Union, List, Any

import pandas as pd

# import your validator (assumes validate_dataset and HardValidationError live in validate_dataset.py)
from .validation import validate_dataset, normalize_col, HardValidationError


def load_json(path: str) -> Any:
    """Load JSON from a file, returning {} on failure."""
    try:
        with open(path, "r") as f:
            return json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        return {}


def infer_column_schema(series: pd.Series, cate_threshold: int = 10) -> Any:
    """
    Infer a minimal Pandera‐style schema for a pandas Series.
    """
    import numpy as np

    non_null = series.dropna()
    uniques = non_null.unique().tolist()
    # small categorical
    if 1 < len(uniques) <= cate_threshold:
        cats = [(v.item() if isinstance(v, np.generic) else v) for v in uniques]
        return {
            "dtype": "category",
            "categories": cats,
            "coerce": True,
            "nullable": bool(series.isna().any()),
            "required": True,
            "aliases": [],
            "checks": [{"type": "isin", "args": [cats]}],
        }

    # numeric / datetime / bool / string fallback
    dt = series.dtype
    if pd.api.types.is_integer_dtype(dt):
        dtype = "float64"
        checks = [{"type": "ge", "args": [0]}]
    elif pd.api.types.is_float_dtype(dt):
        dtype = "float64"
        checks = [{"type": "ge", "args": [0.0]}]
    elif pd.api.types.is_bool_dtype(dt):
        dtype = "boolean"
        checks = []
    elif pd.api.types.is_datetime64_any_dtype(dt):
        dtype = "datetime64[ns]"
        checks = [{"type": "not_null", "args": []}]
    else:
        dtype = "string"
        checks = []

    return {
        "dtype": dtype,
        "coerce": True,
        "nullable": bool(series.isna().any()),
        "required": True,
        "aliases": [],
        "checks": checks,
    }


def generate_extension_schema(
    df: Union[pd.DataFrame, str], models: Union[str, List[str]], institution_id: str
) -> Any:
    """
    1) run validate_dataset(...) to detect hard errors or extra_columns
    2) infer schema for each extra column (for extension)
    3) load or init the institution extension JSON
    4) write inferred specs under each model's .data_models[model].columns
    """
    # load & normalize DataFrame
    if isinstance(df, str):
        df = pd.read_csv(df)
    df = df.rename(columns=lambda c: normalize_col(c))

    # run validation; catch HardValidationError to extract extras
    base_schema_path = "'/Volumes/staging_sst_01/default/schema/base_schema.json'"
    try:
        # no hard errors (i.e. no missing_required, extra_columns, or schema errors)
        _ = validate_dataset(df, models, institution_id)
        extras: List[str] = []
    except HardValidationError as e:
        # if missing_required or schema_errors, cannot proceed
        if e.missing_required or e.schema_errors:
            print("Validation FAILED:")
            print(str(e))
            sys.exit(1)
        # otherwise, treat extra_columns as the ones to extend
        extras = e.extra_columns or []

    if not extras:
        print(f"No extra columns to extend for models={models!r}")
        return {}

    # keep only extras actually in df
    extras = [c for c in extras if c in df.columns]
    print(f"Will infer {len(extras)} extra columns: {extras}")

    # infer a spec for each extra column
    inferred = {col: infer_column_schema(df[col]) for col in extras}

    # prepare extension JSON
    output_dir = f"/Volumes/staging_sst_01/{institution_id}_bronze/schema/"
    os.makedirs(output_dir, exist_ok=True)
    inst_fn = f"{institution_id}_schema_extension.json"
    out_path = os.path.join(output_dir, inst_fn)

    extension = load_json(out_path)
    if not extension:
        base_version = load_json(base_schema_path).get("version", "1.0.0")
        extension = {
            "version": base_version,
            "institutions": {institution_id: {"data_models": {}}},
        }

    inst_block = (
        extension.setdefault("institutions", {})
        .setdefault(institution_id, {})
        .setdefault("data_models", {})
    )

    # ensure models is a list
    if isinstance(models, str):
        model_list = [models]
    else:
        model_list = models

    # insert inferred extras under each requested model
    for model in model_list:
        cols_block = inst_block.setdefault(model, {}).setdefault("columns", {})
        cols_block.update(inferred)

    # write back
    with open(out_path, "w") as f:
        json.dump(extension, f, indent=2)

    print(f"Wrote/updated extension schema to {out_path}")
    return extension



import logging
import re
import typing as t
import mlflow

import numpy as np
import numpy.typing as npt
import pandas as pd
import matplotlib.pyplot as plt
import sklearn.base
import shap
from shap import KernelExplainer

LOGGER = logging.getLogger(__name__)


def predict_probs(
    features: pd.DataFrame | np.ndarray,
    model: sklearn.base.BaseEstimator,
    *,
    feature_names: t.Optional[list[str]] = None,
    pos_label: t.Optional[bool | str] = None,
    dtypes: t.Optional[dict[str, object]] = None,
) -> np.ndarray:
    """
    Predict target probabilities for examples in ``features`` using ``model`` .

    Args:
        features
        model
        feature_names: Names of features corresponding to each column in ``features`` ,
            in cases where it must be passed as a numpy array. If not specified,
            feature names are inferred from the model's "column_selector"
            (a standard in models trained using Databricks AutoML).
        pos_label: Value in ``model`` classes that constitutes a "positive" prediction,
            often ``True`` in the case of binary classification.
        dtypes: Mapping of column name to dtype in ``featuress`` that needs to be overridden
            before passing it into ``model`` .

    Returns:
        Predicted probabilities, as a 1-dimensional array (when ``pos_label is True`` )
            or an N-dimensional array, where N corresponds to the number of pred classes.
    """
    if not sklearn.base.is_classifier(model):
        LOGGER.warning("predict_proba() expects a classifier, but received %s", model)

    if feature_names is None:
        feature_names = model.named_steps["column_selector"].get_params()["cols"]
    assert isinstance(feature_names, list)  # type guard

    assert features.shape[1] == len(feature_names)
    if not isinstance(features, pd.DataFrame):
        features = pd.DataFrame(data=features, columns=feature_names)

    if dtypes:
        features = features.astype(dtypes)

    pred_probs = model.predict_proba(features)
    assert isinstance(pred_probs, np.ndarray)  # type guard

    if pos_label is not None:
        return pred_probs[:, model.classes_.tolist().index(pos_label)]
    else:
        return pred_probs


def shap_summary_plot(
    df_shap_values: pd.DataFrame,
    df_test: pd.DataFrame,
    model_feature_names: list[str],
    model_classes: npt.NDArray,
    max_display: int = 20,
) -> None:
    """
    Generates and logs a SHAP summary plot using metadata config.

    Args:
        df_shap_values: DataFrame of SHAP values
        df_test: DataFrame of test features
        model_feature_names: List of feature names used in the model
        model_classes: Numpy array of model classes (e.g. True, False)
        max_display: Maximum number of features to display
    """

    shap.summary_plot(
        df_shap_values.loc[:, model_feature_names].to_numpy(),
        df_test.loc[:, model_feature_names],
        class_names=model_classes,
        max_display=20,
        show=False,
    )

    shap_fig = plt.gcf()

    mlflow.log_figure(shap_fig, "feature_importances_by_shap_plot.png")


def select_top_features_for_display(
    features: pd.DataFrame,
    unique_ids: pd.Series,
    predicted_probabilities: list[float],
    shap_values: npt.NDArray[np.float64],
    n_features: int = 3,
    needs_support_threshold_prob: t.Optional[float] = 0.5,
    features_table: t.Optional[dict[str, dict[str, str]]] = None,
) -> pd.DataFrame:
    """
    Select most important features from SHAP for each student
    and format for display

    Args:
        features: features used in modeling
        unique_ids: student IDs, of length ``features.shape[0]``
        predicted_probabilities: predicted probabilities for each student, in the same
            order as unique_ids, of shape len(unique_ids)
        shap_values: array of arrays of SHAP values, of shape len(unique_ids)
        n_features: number of important features to return
        needs_support_threshold_prob: Minimum probability in [0.0, 1.0] used to compute
            a boolean "needs support" field added to output records. Values in
            ``predicted_probabilities`` greater than or equal to this threshold result in
            a True value, otherwise it's False; if this threshold is set to null,
            then no "needs support" values are added to the output records.
            Note that this doesn't have to be the "optimal" decision threshold for
            the trained model that produced ``predicted_probabilities`` , it can
            be tailored to a school's preferences and use case.
        features_table: Optional mapping of column to human-friendly feature name/desc,
            loaded via :func:`utils.load_features_table()`

    Returns:
        explainability dataframe for display

    TODO: refactor this functionality so it's vectorized and aggregates by student
    """
    pred_probs = np.asarray(predicted_probabilities)

    top_features_info = []
    for i, (unique_id, predicted_proba) in enumerate(zip(unique_ids, pred_probs)):
        instance_shap_values = shap_values[i]
        top_indices = np.argsort(-np.abs(instance_shap_values))[:n_features]
        top_features = features.columns[top_indices]
        top_feature_values = features.iloc[i][top_features]
        top_shap_values = instance_shap_values[top_indices]

        student_output = {
            "Student ID": unique_id,
            "Support Score": predicted_proba,
        }
        if needs_support_threshold_prob is not None:
            student_output["Support Needed"] = (
                predicted_proba >= needs_support_threshold_prob
            )

        for feature_rank, (feature, feature_value, shap_value) in enumerate(
            zip(top_features, top_feature_values, top_shap_values), start=1
        ):
            feature_name = (
                _get_mapped_feature_name(feature, features_table)
                if features_table is not None
                else feature
            )
            feature_value = (
                str(round(feature_value, 2))
                if isinstance(feature_value, float)
                else str(feature_value)
            )
            student_output |= {
                f"Feature_{feature_rank}_Name": feature_name,
                f"Feature_{feature_rank}_Value": feature_value,
                f"Feature_{feature_rank}_Importance": round(shap_value, 2),
            }

        top_features_info.append(student_output)
    return pd.DataFrame(top_features_info)


def generate_ranked_feature_table(
    features: pd.DataFrame,
    shap_values: npt.NDArray[np.float64],
    features_table: t.Optional[dict[str, dict[str, str]]] = None,
) -> pd.DataFrame:
    """
    Creates a table of all selected features of the model ranked
    by average SHAP magnitude (aka feature importance). We utilize average
    SHAP magnitude & an absolute value because it removes directionality
    from the SHAP values and focuses specifically on importance. This table
    is used in the model cards to provide a comprehensive summary of the model's
    features.

    Args:
        features: feature data used in modeling where columns are the feature
            column names
        shap_values: array of arrays of SHAP values, of shape len(unique_ids)
        features_table: Optional mapping of column to human-friendly feature name/desc,
            loaded via :func:`utils.load_features_table()`

    Returns:
        A ranked pandas DataFrame by average shap magnitude
    """
    feature_metadata = []

    for idx, feature in enumerate(features.columns):
        feature_name = (
            _get_mapped_feature_name(feature, features_table)
            if features_table is not None
            else feature
        )
        dtype = features[feature].dtype
        data_type = (
            "Boolean"
            if pd.api.types.is_bool_dtype(dtype)
            else "Continuous"
            if pd.api.types.is_numeric_dtype(dtype)
            else "Categorical"
        )
        avg_shap_magnitude_raw = np.mean(np.abs(shap_values[:, idx]))
        feature_metadata.append(
            {
                "Feature Name": feature_name,
                "Data Type": data_type,
                "Average SHAP Magnitude (Raw)": avg_shap_magnitude_raw,
            }
        )

    df = (
        pd.DataFrame(feature_metadata)
        .sort_values(by="Average SHAP Magnitude (Raw)", ascending=False)
        .reset_index(drop=True)
    )

    # Format magnitudes after sorting to avoid type issues
    df["Average SHAP Magnitude"] = df["Average SHAP Magnitude (Raw)"].apply(
        lambda x: "<0.0000" if round(x, 4) == 0 else round(x, 4)
    )

    # Drop the raw magnitude column
    df = df.drop(columns=["Average SHAP Magnitude (Raw)"])

    # Log as an ML artifact
    df.to_csv("/tmp/ranked_selected_features.csv", index=False)
    mlflow.log_artifact(
        "/tmp/ranked_selected_features.csv", artifact_path="selected_features"
    )

    return df


def _get_mapped_feature_name(
    feature_col: str, features_table: dict[str, dict[str, str]], metadata: bool = False
) -> t.Any:
    feature_col = feature_col.lower()  # just in case
    if feature_col in features_table:
        entry = features_table[feature_col]
        feature_name = entry["name"]
        if metadata:
            short_desc = entry.get("short_desc")
            long_desc = entry.get("long_desc")
            return feature_name, short_desc, long_desc
        return feature_name
    else:
        for fkey, fval in features_table.items():
            if "(" in fkey and ")" in fkey:
                if match := re.fullmatch(fkey, feature_col):
                    feature_name = fval["name"].format(*match.groups())
                    if metadata:
                        short_desc = fval.get("short_desc")
                        long_desc = fval.get("long_desc")
                        return feature_name, short_desc, long_desc
                    return feature_name

        else:
            feature_name = feature_col
            if metadata:
                return feature_name, None, None
            return feature_name


def calculate_shap_values_spark_udf(
    dfs: t.Iterator[pd.DataFrame],
    *,
    student_id_col: str,
    model_features: list[str],
    explainer: KernelExplainer,
    mode: pd.Series,
) -> t.Iterator[pd.DataFrame]:
    """
    SHAP is computationally expensive, so this function enables parallelization,
    by calculating SHAP values over an iterator of DataFrames. Sparks' repartition
    performs a full shuffle (does not preserve row order), so it is critical to
    extract the student_id_col prior to creating shap values and then reattach
    for our final output.

    Args:
        dfs: An iterator over Pandas DataFrames.
        Each DataFrame is a batch of data points.
        student_id_col: The name of the column containing student_id
        model_features: A list of strings representing the names
        of the features for our model
        explainer: A KernelExplainer object used to compute
        shap values from our loaded model.
        mode: A Series containing values to impute missing values

    Returns:
        Iterator[pd.DataFrame]: An iterator over Pandas DataFrames. Each DataFrame
        contains the SHAP values for that partition of data.
    """
    for df in dfs:
        yield calculate_shap_values(
            df,
            explainer,
            feature_names=model_features,
            fillna_values=mode,
            student_id_col=student_id_col,
        )


def calculate_shap_values(
    df: pd.DataFrame,
    explainer: KernelExplainer,
    *,
    feature_names: list[str],
    fillna_values: pd.Series,
    student_id_col: str = "student_id",
) -> pd.DataFrame:
    """
    Compute SHAP values for the features in ``df`` using ``explainer`` and return result
    as a reassembled data frame with ``feature_names`` as columns and ``student_id_col``
    added as an extra (identifying) column.

    Args:
        df
        explainer
        feature_names
        fillna_values
        student_id_col

    Reference:
        https://shap.readthedocs.io/en/stable/generated/shap.KernelExplainer.html
    """
    # preserve student ids
    student_ids = df.loc[:, student_id_col].reset_index(drop=True)
    # impute missing values and run shap values using just features
    features_imp = df.loc[:, feature_names].fillna(fillna_values)
    explanation = explainer(features_imp)
    return (
        pd.DataFrame(data=explanation.values, columns=explanation.feature_names)
        # reattach student ids to their shap values
        .assign(**{student_id_col: student_ids})
    )


def top_shap_features(
    features: pd.DataFrame,
    unique_ids: pd.Series,
    shap_values: npt.NDArray[np.float64],
    top_n: int = 10,
    features_table: t.Optional[dict[str, dict[str, str]]] = None,
) -> pd.DataFrame:
    """
    Extracts the top N most important SHAP features across all samples.

    Args:
        features (pd.DataFrame): Input feature values.
        unique_ids (pd.Series): Unique identifiers for each sample.
        shap_values (np.ndarray): SHAP values for the input features.
        top_n (int): Number of top features to select (default is 10).
        features_table (dict, optional): Mapping of feature names to human-readable names.

    Returns:
        pd.DataFrame: Long-form DataFrame with columns:
            - student_id
            - feature_name
            - shap_value
            - feature_value
    """

    if features.empty or shap_values.size == 0 or unique_ids.empty:
        raise ValueError("Input data cannot be empty.")

    shap_long = (
        pd.DataFrame(shap_values, columns=features.columns)
        .assign(student_id=unique_ids.values)
        .melt(id_vars="student_id", var_name="feature_name", value_name="shap_value")
    )

    feature_long = features.assign(student_id=unique_ids.values).melt(
        id_vars="student_id", var_name="feature_name", value_name="feature_value"
    )

    summary_df = shap_long.merge(feature_long, on=["student_id", "feature_name"])

    top_n_features = (
        summary_df.groupby("feature_name")["shap_value"]
        .apply(lambda x: np.mean(np.abs(x)))
        .sort_values(ascending=False)
        .head(top_n)
        .index.tolist()
    )

    top_features = summary_df[summary_df["feature_name"].isin(top_n_features)].copy()

    if features_table is not None:
        top_features[
            ["feature_readable_name", "feature_short_desc", "feature_long_desc"]
        ] = top_features["feature_name"].apply(
            lambda feature: pd.Series(
                _get_mapped_feature_name(feature, features_table, metadata=True)
            )
        )

    top_features["feature_value"] = top_features["feature_value"].astype(str)

    return top_features


def support_score_distribution_table(
    df_serving: pd.DataFrame,
    unique_ids: t.Any,
    pred_probs: t.Any,
    shap_values: t.Any,
    inference_params: dict,
    features_table: t.Optional[dict[str, dict[str, str]]] = None,
) -> pd.DataFrame:
    """
    Selects top SHAP features for each student, and bins the support scores.

    Args:
        df_serving (pd.DataFrame): Input features used for prediction.
        unique_ids (pd.Series): Unique ids (student_id) for each student.
        pred_probs (list or np.ndarray): Predicted probabilities from the model.
        shap_values (np.ndarray or pd.DataFrame): SHAP values for the input features.
        inference_params (dict): Dictionary containing configuration for:
            - "num_top_features" (int): Number of top features to display.
            - "min_prob_pos_label" (float): Threshold to determine if support is needed.
        features_table (dict): Optional dictionary mapping feature names to understandable format.
        model_feature_names (list): List of feature names used by the model.

    Returns:
        pd.DataFrame: A DataFrame with the following columns:
            - bin_lower: Lower bound of the support score bin.
            - bin_upper: Upper bound of the support score bin.
            - support_score: Midpoint of the bin (used for plotting).
            - count_of_students: Number of students in the bin.
            - pct: Percentage of total students in the bin.

    """

    try:
        result = select_top_features_for_display(
            features=df_serving,
            unique_ids=unique_ids,
            predicted_probabilities=pred_probs,
            shap_values=shap_values.values,
            n_features=inference_params["num_top_features"],
            needs_support_threshold_prob=inference_params["min_prob_pos_label"],
            features_table=features_table,
        )

        # --- Bin support scores for histogram (e.g., 0.0 to 1.0 in 0.1 steps) ---
        bins = np.arange(0.1, 1.1, 0.1)
        result["score_bin"] = pd.cut(
            result["Support Score"], bins=bins, include_lowest=True, right=False
        )

        # Group and count
        bin_counts = (
            result.groupby("score_bin", observed=True)
            .size()
            .reset_index(name="count_of_students")
        )

        # Extract bin boundaries
        bin_counts["bin_lower"] = bin_counts["score_bin"].apply(
            lambda x: round(x.left, 2)
        )
        bin_counts["bin_upper"] = bin_counts["score_bin"].apply(
            lambda x: round(x.right, 2)
        )
        bin_counts["support_score"] = bin_counts["score_bin"].apply(
            lambda x: round((x.left + x.right) / 2, 2)
        )

        total_students = len(result)
        bin_counts["pct"] = (
            bin_counts["count_of_students"] / total_students * 100
        ).round(2)

        return bin_counts[
            ["bin_lower", "bin_upper", "support_score", "count_of_students", "pct"]
        ]

    except Exception:
        import traceback

        traceback.print_exc()
        raise  # <-- temporarily raise instead of returning None



import logging

import pandas as pd

LOGGER = logging.getLogger(__name__)


def dedupe_by_renumbering_courses(df: pd.DataFrame) -> pd.DataFrame:
    """
    Deduplicate rows in raw course data ``df`` by renumbering courses, such that
    the data passes data schema uniqueness requirements.

    Args:
        df: Raw course dataset

    Warning:
        This logic assumes that all rows are actually valid, and that the school's
        course numbering is wonky. This can happen if, for example, supplemental courses
        (e.g. labs) for a main course share the same course number.

        Don't use this function if there are actual duplicate records in the data!

    See Also:
        :class:`schemas.pdp.RawPDPCourseDataSchema`
    """
    # HACK: infer the correct student id col in raw data from the data itself
    student_id_col = (
        "student_guid"
        if "student_guid" in df.columns
        else "study_id"
        if "study_id" in df.columns
        else "student_id"
    )
    unique_cols = [
        student_id_col,
        "academic_year",
        "academic_term",
        "course_prefix",
        "course_number",
        "section_id",
    ]
    deduped_course_numbers = (
        # get all duplicated records
        df.loc[df.duplicated(unique_cols, keep=False), :]
        # sort in descending order of num credits attempted, which assumes that
        # courses worth more credits are "primary" and should be numbered first
        .sort_values(
            by=unique_cols + ["number_of_credits_attempted"],
            ascending=False,
            ignore_index=False,
        )
        .assign(
            grp_num=(
                lambda df: df.groupby(unique_cols)["course_number"].transform(
                    "cumcount"
                )
                + 1
            ),
            # add "group number" suffix to course numbers in order to disambiguate
            course_number=lambda df: df["course_number"].str.cat(
                df["grp_num"].astype("string"), sep="-"
            ),
        )
        .loc[:, ["course_number"]]
    )
    LOGGER.warning(
        "%s duplicate course records found; course numbers modified to avoid duplicates",
        len(deduped_course_numbers),
    )
    # update course numbers for dupe'd records in-place
    df.update(deduped_course_numbers, overwrite=True)
    return df



import pandas as pd
import pytest

from student_success_tool.dataio.converters import pdp


@pytest.fixture(scope="module")
def df_test():
    return pd.DataFrame(
        {
            "student_id": ["01", "01", "01", "01", "02", "02"],
            "academic_year": [
                "2024-25",
                "2024-25",
                "2024-25",
                "2024-25",
                "2020-21",
                "2020-21",
            ],
            "academic_term": ["FALL", "FALL", "FALL", "SPRING", "SUMMER", "SUMMER"],
            "course_prefix": ["PHYS", "PHYS", "MATH", "MATH", "ENGL", "ENGL"],
            "course_number": ["101", "101", "101", "202", "102", "102"],
            "section_id": ["001", "001", "123", "123", "456", "456"],
            "number_of_credits_attempted": [1.0, 3.0, 3.0, 3.0, 4.0, 1.0],
        },
        dtype="string",
    ).astype({"number_of_credits_attempted": "Float32"})


@pytest.mark.parametrize(
    "exp",
    [
        pd.DataFrame(
            {
                "student_id": ["01", "01", "01", "01", "02", "02"],
                "academic_year": [
                    "2024-25",
                    "2024-25",
                    "2024-25",
                    "2024-25",
                    "2020-21",
                    "2020-21",
                ],
                "academic_term": ["FALL", "FALL", "FALL", "SPRING", "SUMMER", "SUMMER"],
                "course_prefix": ["PHYS", "PHYS", "MATH", "MATH", "ENGL", "ENGL"],
                "course_number": ["101-2", "101-1", "101", "202", "102-1", "102-2"],
                "section_id": ["001", "001", "123", "123", "456", "456"],
                "number_of_credits_attempted": [1.0, 3.0, 3.0, 3.0, 4.0, 1.0],
            },
            dtype="string",
        ).astype({"number_of_credits_attempted": "Float32"})
    ],
)
def test_dedupe_by_renumbering_courses(df_test, exp):
    obs = pdp.dedupe_by_renumbering_courses(df_test)
    assert isinstance(obs, pd.DataFrame)
    assert pd.testing.assert_frame_equal(obs, exp) is None



import pandas as pd
import pytest

from student_success_tool.preprocessing import checkpoints
from student_success_tool.preprocessing.targets.pdp import shared


@pytest.fixture(scope="module")
def df_test_max_term():
    return pd.DataFrame(
        data={
            "student_id": ["01", "01", "02", "02", "03"],
            "enrollment_intensity": ["FT", "FT", "PT", "PT", "FT"],
            "term_rank": [1, 2, 4, 5, 10],
            "term_is_pre_cohort": [False, False, True, False, False],
        },
    ).astype({"student_id": "string", "enrollment_intensity": "string"})


@pytest.fixture(scope="module")
def df_test_year2():
    return pd.DataFrame(
        {
            "student_id": ["01", "01", "02", "03"],
            "cohort_id": [
                "2020-21 FALL",
                "2020-21 FALL",
                "2021-22 SPRING",
                "2022-23 FALL",
            ],
            "term_id": [
                "2020-21 FALL",
                "2020-21 SPRING",
                "2021-22 SPRING",
                "2022-23 FALL",
            ],
        },
        dtype="string",
    )


@pytest.mark.parametrize(
    [
        "checkpoint",
        "intensity_time_limits",
        "max_term_rank",
        "num_terms_in_year",
        "exp",
    ],
    [
        # max target term inferred
        (
            checkpoints.pdp.first_student_terms,
            {"FT": (1, "year"), "PT": (2, "year")},
            "infer",
            3,
            pd.DataFrame(
                data={"student_max_term_rank": [4, 10], "max_term_rank": [10, 10]},
                index=pd.Index(["01", "02"], name="student_id", dtype="string"),
            ).astype("Int8"),
        ),
        # max target term manually specified
        (
            checkpoints.pdp.first_student_terms,
            {"FT": (1, "year"), "PT": (2, "year")},
            13,
            3,
            pd.DataFrame(
                data={
                    "student_max_term_rank": [4, 10, 13],
                    "max_term_rank": [13, 13, 13],
                },
                index=pd.Index(["01", "02", "03"], name="student_id", dtype="string"),
            ).astype("Int8"),
        ),
        # num terms in year adjusted
        (
            checkpoints.pdp.first_student_terms,
            {"FT": (1, "year"), "PT": (2, "year")},
            "infer",
            4,
            pd.DataFrame(
                data={"student_max_term_rank": [5], "max_term_rank": [10]},
                index=pd.Index(["01"], name="student_id", dtype="string"),
            ).astype("Int8"),
        ),
        # checkpoint given as dataframe
        (
            pd.DataFrame(
                data={
                    "student_id": ["01", "02", "03"],
                    "enrollment_intensity": ["FT", "PT", "FT"],
                    "term_rank": [1, 4, 10],
                    "term_is_pre_cohort": [False, True, False],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"FT": (1, "year"), "PT": (2, "year")},
            13,
            3,
            pd.DataFrame(
                data={
                    "student_max_term_rank": [4, 10, 13],
                    "max_term_rank": [13, 13, 13],
                },
                index=pd.Index(["01", "02", "03"], name="student_id", dtype="string"),
            ).astype("Int8"),
        ),
        # pre-cohort terms excluded when computing max target term
        (
            checkpoints.pdp.first_student_terms_within_cohort,
            {"FT": (1, "year"), "PT": (2, "year")},
            12,
            3,
            pd.DataFrame(
                data={"student_max_term_rank": [4, 11], "max_term_rank": [12, 12]},
                index=pd.Index(["01", "02"], name="student_id", dtype="string"),
            ).astype("Int8"),
        ),
    ],
)
def test_get_students_with_max_target_term_in_dataset(
    df_test_max_term,
    checkpoint,
    intensity_time_limits,
    max_term_rank,
    num_terms_in_year,
    exp,
):
    obs = shared.get_students_with_max_target_term_in_dataset(
        df_test_max_term,
        checkpoint=checkpoint,
        intensity_time_limits=intensity_time_limits,
        max_term_rank=max_term_rank,
        num_terms_in_year=num_terms_in_year,
        student_id_cols="student_id",
        enrollment_intensity_col="enrollment_intensity",
        term_rank_col="term_rank",
    )
    assert isinstance(obs, pd.DataFrame)
    assert pd.testing.assert_frame_equal(obs, exp) is None


@pytest.mark.parametrize(
    ["max_academic_year", "exp"],
    [
        (
            "infer",
            pd.DataFrame(
                data={
                    "student_cohort_year": [2020, 2021],
                    "max_academic_year": [2022, 2022],
                },
                index=pd.Index(["01", "02"], name="student_id", dtype="string"),
                dtype="Int32",
            ),
        ),
        (
            "2024-25",
            pd.DataFrame(
                data={
                    "student_cohort_year": [2020, 2021, 2022],
                    "max_academic_year": [2024, 2024, 2024],
                },
                index=pd.Index(["01", "02", "03"], name="student_id", dtype="string"),
                dtype="Int32",
            ),
        ),
    ],
)
def test_get_students_with_second_year_in_dataset(
    df_test_year2, max_academic_year, exp
):
    obs = shared.get_students_with_second_year_in_dataset(
        df_test_year2,
        max_academic_year=max_academic_year,
        student_id_cols="student_id",
        cohort_id_col="cohort_id",
        term_id_col="term_id",
    )
    assert isinstance(obs, pd.DataFrame)
    assert pd.testing.assert_frame_equal(obs, exp) is None



import pandas as pd
import pytest

from student_success_tool.preprocessing.checkpoints import pdp


@pytest.fixture(scope="module")
def df_test():
    return pd.DataFrame(
        {
            "student_id": [
                "01",
                "01",
                "01",
                "01",
                "02",
                "02",
                "02",
                "03",
                "03",
                "04",
                "05",
            ],
            "term_rank": [3, 4, 5, 6, 5, 6, 7, 2, 3, 4, 8],
            "cohort_id": [
                "2020-21 SPRING",
                "2020-21 SPRING",
                "2020-21 SPRING",
                "2020-21 SPRING",
                "2021-22 FALL",
                "2021-22 FALL",
                "2021-22 FALL",
                "2019-20 FALL",
                "2019-20 FALL",
                "2020-21 SPRING",
                "2022-23 FALL",
            ],
            "term_id": [
                "2020-21 FALL",
                "2020-21 SPRING",
                "2021-22 FALL",
                "2021-22 WINTER",
                "2021-22 FALL",
                "2023-24 FALL",
                "2024-25 SUMMER",
                "2019-20 SPRING",
                "2019-20 WINTER",
                "2020-21 SPRING",
                "2022-23 FALL",
            ],
            "enrollment_year": [1, 1, 2, 2, 1, 3, 4, 1, 1, 1, 1],
            "enrollment_intensity": [
                "FULL-TIME",
                "FULL-TIME",
                "FULL-TIME",
                "FULL-TIME",
                "PART-TIME",
                "PART-TIME",
                "PART-TIME",
                "PART-TIME",
                "PART-TIME",
                "FULL-TIME",
                pd.NA,
            ],
            "num_credits_earned": [
                25.0,
                30.0,
                35.0,
                12.0,
                25.0,
                35.0,
                18.0,
                20.0,
                5.0,
                45.0,
                10.0,
            ],
            "term_is_pre_cohort": [
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            "term_is_core": [
                True,
                True,
                True,
                False,
                True,
                True,
                False,
                True,
                False,
                True,
                True,
            ],
        },
    ).astype(
        {
            "student_id": "string",
            "cohort_id": "string",
            "term_id": "string",
            "enrollment_intensity": "string",
        }
    )


@pytest.mark.parametrize(
    [
        "n",
        "include_cols",
        "exclude_pre_cohort_terms",
        "exclude_non_core_terms",
        "enrollment_year_col",
        "valid_enrollment_year",
        "exp",
    ],
    [
        (
            0,
            None,
            False,
            False,
            "enrollment_year",
            1,
            pd.DataFrame(
                {
                    "student_id": ["01", "02", "03", "04", "05"],
                    "term_rank": [3, 5, 2, 4, 8],
                    "cohort_id": [
                        "2020-21 SPRING",
                        "2021-22 FALL",
                        "2019-20 FALL",
                        "2020-21 SPRING",
                        "2022-23 FALL",
                    ],
                    "term_id": [
                        "2020-21 FALL",
                        "2021-22 FALL",
                        "2019-20 SPRING",
                        "2020-21 SPRING",
                        "2022-23 FALL",
                    ],
                    "enrollment_year": [1, 1, 1, 1, 1],
                    "enrollment_intensity": [
                        "FULL-TIME",
                        "PART-TIME",
                        "PART-TIME",
                        "FULL-TIME",
                        pd.NA,
                    ],
                    "num_credits_earned": [25.0, 25.0, 20.0, 45.0, 10.0],
                    "term_is_pre_cohort": [True, False, False, False, False],
                    "term_is_core": [True, True, True, True, True],
                }
            ).astype(
                {
                    "student_id": "string",
                    "cohort_id": "string",
                    "term_id": "string",
                    "enrollment_intensity": "string",
                }
            ),
        ),
        (
            0,
            ["term_id"],
            True,
            True,
            None,
            None,
            pd.DataFrame(
                {
                    "student_id": ["01", "02", "03", "04", "05"],
                    "term_rank": [4, 5, 2, 4, 8],
                    "term_id": [
                        "2020-21 SPRING",
                        "2021-22 FALL",
                        "2019-20 SPRING",
                        "2020-21 SPRING",
                        "2022-23 FALL",
                    ],
                }
            ).astype({"student_id": "string", "term_id": "string"}),
        ),
        (
            1,
            None,
            True,
            True,
            "enrollment_year",
            3,
            pd.DataFrame(
                {
                    "student_id": ["02"],
                    "term_rank": [6],
                    "cohort_id": ["2021-22 FALL"],
                    "term_id": ["2023-24 FALL"],
                    "enrollment_year": [3],
                    "enrollment_intensity": ["PART-TIME"],
                    "num_credits_earned": [35.0],
                    "term_is_pre_cohort": [False],
                    "term_is_core": [True],
                }
            ).astype(
                {
                    "student_id": "string",
                    "cohort_id": "string",
                    "term_id": "string",
                    "enrollment_intensity": "string",
                }
            ),
        ),
        (
            -1,
            None,
            False,
            False,
            None,
            None,
            pd.DataFrame(
                {
                    "student_id": ["01", "02", "03", "04", "05"],
                    "term_rank": [6, 7, 3, 4, 8],
                    "cohort_id": [
                        "2020-21 SPRING",
                        "2021-22 FALL",
                        "2019-20 FALL",
                        "2020-21 SPRING",
                        "2022-23 FALL",
                    ],
                    "term_id": [
                        "2021-22 WINTER",
                        "2024-25 SUMMER",
                        "2019-20 WINTER",
                        "2020-21 SPRING",
                        "2022-23 FALL",
                    ],
                    "enrollment_year": [2, 4, 1, 1, 1],
                    "enrollment_intensity": [
                        "FULL-TIME",
                        "PART-TIME",
                        "PART-TIME",
                        "FULL-TIME",
                        pd.NA,
                    ],
                    "num_credits_earned": [12.0, 18.0, 5.0, 45.0, 10.0],
                    "term_is_pre_cohort": [False, False, False, False, False],
                    "term_is_core": [False, False, False, True, True],
                }
            ).astype(
                {
                    "student_id": "string",
                    "cohort_id": "string",
                    "term_id": "string",
                    "enrollment_intensity": "string",
                }
            ),
        ),
    ],
)
def test_nth_student_terms(
    df_test,
    n,
    include_cols,
    exclude_pre_cohort_terms,
    exclude_non_core_terms,
    enrollment_year_col,
    valid_enrollment_year,
    exp,
):
    from student_success_tool.preprocessing.checkpoints import pdp

    obs = pdp.nth_student_terms(
        df_test,
        n=n,
        student_id_cols="student_id",
        sort_cols="term_rank",
        include_cols=include_cols,
        term_is_pre_cohort_col="term_is_pre_cohort",
        exclude_pre_cohort_terms=exclude_pre_cohort_terms,
        term_is_core_col="term_is_core",
        exclude_non_core_terms=exclude_non_core_terms,
        enrollment_year_col=enrollment_year_col,
        valid_enrollment_year=valid_enrollment_year,
    )
    assert isinstance(obs, pd.DataFrame)
    assert (
        pd.testing.assert_frame_equal(
            obs.sort_values("student_id").reset_index(drop=True),
            exp.sort_values("student_id").reset_index(drop=True),
        )
        is None
    )


@pytest.mark.parametrize(
    ["include_cols", "exclude_pre_cohort_terms", "exclude_non_core_terms", "exp"],
    [
        (
            ["term_id"],
            False,
            True,
            pd.DataFrame(
                data={
                    "student_id": ["01", "02", "03", "04", "05"],
                    "term_rank": [3, 5, 2, 4, 8],
                    "term_id": [
                        "2020-21 FALL",
                        "2021-22 FALL",
                        "2019-20 SPRING",
                        "2020-21 SPRING",
                        "2022-23 FALL",
                    ],
                },
                index=pd.Index([0, 3, 5, 6, 7], dtype="int64"),
            ).astype({"student_id": "string", "term_id": "string"}),
        ),
    ],
)
def test_first_student_terms(
    df_test, include_cols, exclude_pre_cohort_terms, exclude_non_core_terms, exp
):
    obs = pdp.first_student_terms(
        df_test,
        student_id_cols="student_id",
        sort_cols="term_rank",
        include_cols=include_cols,
        term_is_pre_cohort_col="term_is_pre_cohort",
        exclude_pre_cohort_terms=exclude_pre_cohort_terms,
        term_is_core_col="term_is_core",
        exclude_non_core_terms=exclude_non_core_terms,
    )
    assert isinstance(obs, pd.DataFrame)
    pd.testing.assert_frame_equal(
        obs.reset_index(drop=True), exp.reset_index(drop=True)
    )


@pytest.mark.parametrize(
    ["include_cols", "exclude_pre_cohort_terms", "exclude_non_core_terms", "exp"],
    [
        (
            ["term_id"],
            False,
            False,
            pd.DataFrame(
                data={
                    "student_id": ["01", "02", "03", "04", "05"],
                    "term_rank": [6, 7, 3, 4, 8],
                    "term_id": [
                        "2021-22 WINTER",
                        "2024-25 SUMMER",
                        "2019-20 WINTER",
                        "2020-21 SPRING",
                        "2022-23 FALL",
                    ],
                },
                index=pd.Index([2, 4, 5, 6, 7], dtype="int64"),
            ).astype({"student_id": "string", "term_id": "string"}),
        ),
    ],
)
def test_last_student_terms(
    df_test, include_cols, exclude_pre_cohort_terms, exclude_non_core_terms, exp
):
    obs = pdp.last_student_terms(
        df_test,
        student_id_cols="student_id",
        sort_cols="term_rank",
        include_cols=include_cols,
        term_is_pre_cohort_col="term_is_pre_cohort",
        exclude_pre_cohort_terms=exclude_pre_cohort_terms,
        term_is_core_col="term_is_core",
        exclude_non_core_terms=exclude_non_core_terms,
    )
    assert isinstance(obs, pd.DataFrame)
    pd.testing.assert_frame_equal(
        obs.reset_index(drop=True), exp.reset_index(drop=True)
    )


@pytest.mark.parametrize(
    ["min_num_credits", "include_cols", "exp"],
    [
        (
            30.0,
            ["num_credits_earned"],
            pd.DataFrame(
                data={
                    "student_id": ["01", "02", "04"],
                    "term_rank": [4, 6, 4],
                    "num_credits_earned": [30.0, 35.0, 45.0],
                },
                index=pd.Index([1, 4, 6], dtype="int64"),
            ).astype({"student_id": "string"}),
        ),
        (
            45.0,
            ["num_credits_earned"],
            pd.DataFrame(
                data={
                    "student_id": ["04"],
                    "term_rank": [4],
                    "num_credits_earned": [45.0],
                },
                index=pd.Index([6], dtype="int64"),
            ).astype({"student_id": "string"}),
        ),
    ],
)
def test_first_student_terms_at_num_credits_earned(
    df_test, min_num_credits, include_cols, exp
):
    obs = pdp.first_student_terms_at_num_credits_earned(
        df_test,
        min_num_credits=min_num_credits,
        num_credits_col="num_credits_earned",
        student_id_cols="student_id",
        sort_cols="term_rank",
        include_cols=include_cols,
    )
    assert isinstance(obs, pd.DataFrame)
    pd.testing.assert_frame_equal(
        obs.reset_index(drop=True), exp.reset_index(drop=True)
    )


@pytest.mark.parametrize(
    ["include_cols", "exp"],
    [
        (
            ["term_is_pre_cohort"],
            pd.DataFrame(
                data={
                    "student_id": ["01", "02", "03", "04", "05"],
                    "term_rank": [4, 5, 2, 4, 8],
                    "term_is_pre_cohort": [False, False, False, False, False],
                },
                index=pd.Index([1, 3, 5, 6, 7], dtype="int64"),
            ).astype({"student_id": "string"}),
        ),
    ],
)
def test_first_student_terms_within_cohort(df_test, include_cols, exp):
    obs = pdp.first_student_terms_within_cohort(
        df_test,
        term_is_pre_cohort_col="term_is_pre_cohort",
        student_id_cols="student_id",
        sort_cols="term_rank",
        include_cols=include_cols,
    )
    assert isinstance(obs, pd.DataFrame)
    pd.testing.assert_frame_equal(
        obs.reset_index(drop=True), exp.reset_index(drop=True)
    )


@pytest.mark.parametrize(
    ["enrollment_year", "include_cols", "exp"],
    [
        (
            1,
            ["enrollment_year"],
            pd.DataFrame(
                data={
                    "student_id": ["01", "02", "03", "04", "05"],
                    "term_rank": [4, 5, 3, 4, 8],
                    "enrollment_year": [1, 1, 1, 1, 1],
                },
                index=pd.Index([1, 4, 8, 9, 10], dtype="int64"),
            ).astype({"student_id": "string"}),
        ),
        (
            2,
            ["enrollment_year"],
            pd.DataFrame(
                data={
                    "student_id": ["01"],
                    "term_rank": [6],
                    "enrollment_year": [2],
                },
                index=pd.Index([3], dtype="int64"),
            ).astype({"student_id": "string"}),
        ),
    ],
)
def test_last_student_terms_in_enrollment_year(
    df_test, enrollment_year, include_cols, exp
):
    obs = pdp.last_student_terms_in_enrollment_year(
        df_test,
        enrollment_year=enrollment_year,
        enrollment_year_col="enrollment_year",
        student_id_cols="student_id",
        sort_cols="term_rank",
        include_cols=include_cols,
    )
    assert isinstance(obs, pd.DataFrame)
    pd.testing.assert_frame_equal(
        obs.reset_index(drop=True), exp.reset_index(drop=True)
    )



import functools as ft
import re
import typing as t
from collections.abc import Sequence
from datetime import date

import pandas as pd

from ....utils import types

RE_YEAR_TERM = re.compile(
    r"(?P<start_yr>\d{4})-(?P<end_yr>\d{2}) (?P<term>FALL|WINTER|SPRING|SUMMER)",
    flags=re.IGNORECASE,
)

TERM_BOUND_MONTH_DAYS = {
    "FALL": {"start": (9, 1), "end": (12, 31)},
    "WINTER": {"start": (1, 1), "end": (1, 31)},
    "SPRING": {"start": (2, 1), "end": (5, 31)},
    "SUMMER": {"start": (6, 1), "end": (8, 31)},
}


def extract_short_cip_code(ser: pd.Series) -> pd.Series:
    # NOTE: this simpler form works, but the values aren't nearly as clean
    # return ser.str.slice(stop=2).str.strip(".")
    return (
        ser.str.extract(r"^(?P<subject_area>\d[\d.])[\d.]+$", expand=False)
        .str.strip(".")
        .astype("string")
    )


def frac_credits_earned(
    df: pd.DataFrame,
    *,
    earned_col: str = "num_credits_earned",
    attempted_col: str = "num_credits_attempted",
) -> pd.Series:
    return df[earned_col].div(df[attempted_col])


def compute_values_equal(ser: pd.Series, to: t.Any | list[t.Any]) -> pd.Series:
    return ser.isin(to) if isinstance(to, list) else ser.eq(to)


def merge_many_dataframes(
    dfs: Sequence[pd.DataFrame],
    *,
    on: str | list[str],
    how: t.Literal["left", "right", "outer", "inner"] = "inner",
    sort: bool = False,
) -> pd.DataFrame:
    """
    Merge 2+ dataframes using the same set of merge parameters for each operation.

    See Also:
        - https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.merge.html
    """
    return ft.reduce(
        lambda left, right: pd.merge(left, right, on=on, how=how, sort=sort), dfs
    )


def year_term(df: pd.DataFrame, *, year_col: str, term_col: str) -> pd.Series:
    return df[year_col].str.cat(df[term_col], sep=" ")


def year_term_dt(
    df: pd.DataFrame,
    *,
    col: str,
    bound: t.Literal["start", "end"],
    first_term_of_year: types.TermType,
) -> pd.Series:
    """
    Compute an approximate start/end date for a given year-term,
    e.g. to compute time elapsed between course enrollments or to order course history.

    Args:
        df
        col: Column in ``df`` whose values represent academic/cohort year and term,
            formatted as "YYYY-YY TERM".
        bound: Which bound of the date range spanned by ``year_term`` to return;
            either the start (left) or end (right) bound.
        first_term_of_year: Term that officially begins the institution's academic year,
            either "FALL" or "SUMMER", which determines how the date's year is assigned.

    See Also:
        - :func:`year_term()`
    """
    return (
        df[col]
        .map(
            ft.partial(
                _year_term_to_dt, bound=bound, first_term_of_year=first_term_of_year
            ),
            na_action="ignore",
        )
        .astype("datetime64[s]")
    )


def _year_term_to_dt(
    year_term: str, bound: t.Literal["start", "end"], first_term_of_year: types.TermType
) -> date:
    if match := RE_YEAR_TERM.search(year_term):
        start_yr = int(match["start_yr"])
        term = match["term"].upper()
        yr = (
            start_yr
            if (term == "FALL" or (term == "SUMMER" and first_term_of_year == "SUMMER"))
            else start_yr + 1
        )
        mo, dy = TERM_BOUND_MONTH_DAYS[term][bound]
        return date(yr, mo, dy)
    else:
        raise ValueError(f"invalid year_term value: {year_term}")



import pandas as pd
import pytest

from student_success_tool.dataio.converters import pdp


@pytest.fixture(scope="module")
def df_test():
    return pd.DataFrame(
        {
            "student_id": ["01", "02"],
            "cohort": ["2020-21", "2024-25"],
            "attemptedgatewaymathyear_1": ["Y", "N"],
            "gatewayenglishgradey_1": ["P", "F"],
            "someothermangledcolumn": ["X", "Y"],
        },
        dtype="string",
    )


@pytest.mark.parametrize(
    ["overrides", "exp"],
    [
        (
            None,
            pd.DataFrame(
                {
                    "student_id": ["01", "02"],
                    "cohort": ["2020-21", "2024-25"],
                    "attempted_gateway_math_year_1": ["Y", "N"],
                    "gateway_english_grade_y_1": ["P", "F"],
                    "someothermangledcolumn": ["X", "Y"],
                },
                dtype="string",
            ),
        ),
        (
            {
                "someothermangledcolumn": "some_other_mangled_column",
                "missingcol": "missing_col",
            },
            pd.DataFrame(
                {
                    "student_id": ["01", "02"],
                    "cohort": ["2020-21", "2024-25"],
                    "attempted_gateway_math_year_1": ["Y", "N"],
                    "gateway_english_grade_y_1": ["P", "F"],
                    "some_other_mangled_column": ["X", "Y"],
                },
                dtype="string",
            ),
        ),
    ],
)
def test_rename_mangled_column_names(df_test, overrides, exp):
    obs = pdp.rename_mangled_column_names(df_test, overrides=overrides)
    assert isinstance(obs, pd.DataFrame)
    assert pd.testing.assert_frame_equal(obs, exp) is None



import logging
import typing as t

import mlflow
import mlflow.exceptions
import mlflow.tracking

LOGGER = logging.getLogger(__name__)


def register_mlflow_model(
    model_name: str,
    institution_id: str,
    *,
    run_id: str,
    catalog: str,
    registry_uri: str = "databricks-uc",
    model_alias: t.Optional[str] = "Staging",
    mlflow_client: mlflow.tracking.MlflowClient,
) -> None:
    """
    Register an mlflow model from a run, using run-level tags to prevent duplicate registration.

    Args:
        model_name
        institution_id
        run_id
        catalog
        registry_uri
        model_alias
        mlflow_client

    References:
        - https://mlflow.org/docs/latest/model-registry.html
    """
    model_path = f"{catalog}.{institution_id}_gold.{model_name}"
    mlflow.set_registry_uri(registry_uri)

    # Create the registered model if it doesn't exist
    try:
        mlflow_client.create_registered_model(name=model_path)
        LOGGER.info("New registered model '%s' successfully created", model_path)
    except mlflow.exceptions.MlflowException as e:
        if "RESOURCE_ALREADY_EXISTS" in str(e):
            LOGGER.info("Model '%s' already exists in registry", model_path)
        else:
            raise e

    # Check for the "model_registered" tag on the run
    try:
        run_tags = mlflow_client.get_run(run_id).data.tags
        if run_tags.get("model_registered") == "true":
            LOGGER.info("Run ID '%s' has already been registered. Skipping.", run_id)
            return
    except mlflow.exceptions.MlflowException as e:
        LOGGER.warning("Unable to check tags for run_id '%s': %s", run_id, str(e))
        raise

    # Register new model version
    model_uri = f"runs:/{run_id}/model"
    mv = mlflow_client.create_model_version(
        name=model_path,
        source=model_uri,
        run_id=run_id,
    )
    LOGGER.info("Registered new model version %s from run_id '%s'", mv.version, run_id)

    # Mark the run as registered via tag
    mlflow_client.set_tag(run_id, "model_registered", "true")

    # Optionally assign alias
    if model_alias:
        mlflow_client.set_registered_model_alias(model_path, model_alias, mv.version)
        LOGGER.info("Set alias '%s' to version %s", model_alias, mv.version)


def get_model_name(
    *,
    institution_id: str,
    target: str,
    checkpoint: str,
    extra_info: t.Optional[str] = None,
) -> str:
    """
    Get a standard model name generated from key components, formatted as
    "{institution_id}_{target}_{checkpoint}[_{extra_info}]"
    """
    model_name = f"{institution_id}_{target}_{checkpoint}"
    if extra_info is not None:
        model_name = f"{model_name}_{extra_info}"
    return model_name


def get_mlflow_model_uri(
    *,
    model_name: t.Optional[str] = None,
    model_version: t.Optional[int] = None,
    model_alias: t.Optional[str] = None,
    run_id: t.Optional[str] = None,
    model_path: t.Optional[str] = None,
) -> str:
    """
    Get an mlflow model's URI based on its name, version, alias, path, and/or run id.

    References:
        - https://docs.databricks.com/gcp/en/mlflow/models
        - https://www.mlflow.org/docs/latest/concepts.html#artifact-locations
    """
    if run_id is not None and model_path is not None:
        return f"runs:/{run_id}/{model_path}"
    elif model_name is not None and model_version is not None:
        return f"models:/{model_name}/{model_version}"
    elif model_name is not None and model_alias is not None:
        return f"models:/{model_name}@{model_alias}"
    else:
        raise ValueError(
            "unable to determine model URI from inputs: "
            f"{model_name=}, {model_version=}, {model_alias=}, {model_path=}, {run_id=}"
        )



import faker
import pandas as pd
import pytest

from student_success_tool.dataio.schemas.pdp import RawPDPCohortDataSchema
from student_success_tool.generation.pdp import raw_cohort

FAKER = faker.Faker()
FAKER.add_provider(raw_cohort.Provider)


@pytest.mark.parametrize(
    ["min_cohort_yr", "max_cohort_yr", "normalize_col_names"],
    [(2010, None, False), (2015, 2020, True)],
)
def test_raw_cohort_record(min_cohort_yr, max_cohort_yr, normalize_col_names):
    obs = FAKER.raw_cohort_record(
        min_cohort_yr=min_cohort_yr,
        max_cohort_yr=max_cohort_yr,
        normalize_col_names=normalize_col_names,
    )
    assert obs and isinstance(obs, dict)
    if normalize_col_names is True:
        df_obs = pd.DataFrame([obs])
        obs_valid = RawPDPCohortDataSchema.validate(df_obs, lazy=True)
        assert isinstance(obs_valid, pd.DataFrame)  # => data passed validation
        print(df_obs)


def test_multiple_raw_cohort_records():
    institution_id = 123
    cohort_records = [
        FAKER.raw_cohort_record(normalize_col_names=True, institution_id=institution_id)
        for _ in range(10)
    ]
    df_cohort = pd.DataFrame(cohort_records)
    obs_valid = RawPDPCohortDataSchema.validate(df_cohort, lazy=True)
    assert isinstance(obs_valid, pd.DataFrame)  # => data passed validation



import logging

import pandas as pd

LOGGER = logging.getLogger(__name__)


def add_features(df: pd.DataFrame, *, section_id_cols: list[str]) -> pd.DataFrame:
    """
    Compute section-level features from pdp course dataset w/ added course-level features,
    and add as columns to ``df`` .

    Args:
        df
        section_id_cols: Columns that uniquely identify sections, used to group course rows
            and merge section features back in.

    Note:
        Rows for which any value in ``section_id_cols`` is null won't have features
        computed. This is because such a group is "undefined" in some sense,
        so we can't know if the resulting features are accurate.

    See Also:
        - :func:`pdp.features.course.add_features()`
    """
    LOGGER.info("adding section features ...")
    df_section = (
        df.groupby(by=section_id_cols, as_index=False, observed=True, dropna=True)
        # generating named aggs via functions gives at least *some* testability
        .agg(
            section_num_students_enrolled=section_num_students_enrolled_col_agg(),
            section_num_students_passed=section_num_students_passed_col_agg(),
            section_num_students_completed=section_num_students_completed_col_agg(),
            section_course_grade_numeric_mean=section_course_grade_numeric_mean_col_agg(),
        )
        .astype({"section_course_grade_numeric_mean": "Float32"})
    )
    return pd.merge(df, df_section, on=section_id_cols, how="left")


def section_num_students_enrolled_col_agg(col: str = "student_id") -> pd.NamedAgg:
    return pd.NamedAgg(col, "count")


def section_num_students_passed_col_agg(col: str = "course_passed") -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def section_num_students_completed_col_agg(
    col: str = "course_completed",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "sum")


def section_course_grade_numeric_mean_col_agg(
    col: str = "course_grade_numeric",
) -> pd.NamedAgg:
    return pd.NamedAgg(col, "mean")



import re
import typing as t

import pydantic as pyd

# TODO: set field defaults using literals here instead?
from ..preprocessing.features.pdp import constants
from ..utils import types


class PDPProjectConfig(pyd.BaseModel):
    """Configuration schema for SST PDP projects."""

    institution_id: str = pyd.Field(
        ...,
        description=(
            "Unique (ASCII-only) identifier for institution; used in naming things "
            "such as source directories, catalog schemas, keys in shared configs, etc."
        ),
    )
    institution_name: str = pyd.Field(
        ...,
        description=(
            "Readable 'display' name for institution, distinct from the 'id'; "
            "probably just the school's 'official', public name"
        ),
    )

    # shared parameters
    student_id_col: str = "student_id"
    target_col: str = "target"
    split_col: t.Optional[str] = "split"
    sample_weight_col: t.Optional[str] = "sample_weight"
    student_group_cols: t.Optional[list[str]] = pyd.Field(
        default=["student_age", "race", "ethnicity", "gender", "first_gen"],
        description=(
            "One or more column names in datasets containing student 'groups' "
            "to use for model bias assessment, but *not* as model features"
        ),
    )
    pred_col: str = "pred"
    pred_prob_col: str = "pred_prob"
    pos_label: t.Optional[int | bool | str] = True
    random_state: t.Optional[int] = None

    # key artifacts produced by project pipeline
    datasets: "AllDatasetStagesConfig" = pyd.Field(
        description=(
            "Key datasets produced by the pipeline represented here in this config"
        ),
    )
    model: t.Optional["ModelConfig"] = pyd.Field(
        default=None,
        description=(
            "Essential metadata for identifying and loading trained model artifacts, "
            "as produced by the pipeline represented here in this config"
        ),
    )

    # key steps in project pipeline
    preprocessing: t.Optional["PreprocessingConfig"] = None
    modeling: t.Optional["ModelingConfig"] = None
    inference: t.Optional["InferenceConfig"] = None

    @pyd.computed_field  # type: ignore[misc]
    @property
    def non_feature_cols(self) -> list[str]:
        return (
            [self.student_id_col, self.target_col]
            + ([self.split_col] if self.split_col else [])
            + ([self.sample_weight_col] if self.sample_weight_col else [])
            + (self.student_group_cols or [])
        )

    @pyd.field_validator("institution_id", mode="after")
    @classmethod
    def check_institution_id_isascii(cls, value: str) -> str:
        if not re.search(r"^\w+$", value, flags=re.ASCII):
            raise ValueError(f"institution_id='{value}' is not ASCII-only")
        return value

    # NOTE: this is for *pydantic* model -- not ML model -- configuration
    model_config = pyd.ConfigDict(extra="forbid", strict=True)


class BronzeDatasetConfig(pyd.BaseModel):
    raw_course: "DatasetIOConfig"
    raw_cohort: "DatasetIOConfig"


class SilverDatasetConfig(pyd.BaseModel):
    preprocessed: "DatasetIOConfig"
    modeling: "DatasetIOConfig"


class GoldDatasetConfig(pyd.BaseModel):
    advisor_output: "DatasetIOConfig"


class AllDatasetStagesConfig(pyd.BaseModel):
    bronze: BronzeDatasetConfig
    silver: SilverDatasetConfig
    gold: GoldDatasetConfig


class DatasetIOConfig(pyd.BaseModel):
    table_path: t.Optional[str] = pyd.Field(
        default=None,
        description=(
            "Path to a table in Unity Catalog where dataset is stored, "
            "including the full three-level namespace: 'CATALOG.SCHEMA.TABLE'"
        ),
    )
    file_path: t.Optional[str] = pyd.Field(
        default=None,
        description="Full, absolute path to dataset on disk, e.g. a Databricks Volume",
    )
    # TODO: if/when we allow different file formats, add this parameter ...
    # file_format: t.Optional[t.Literal["csv", "parquet"]] = pyd.Field(default=None)

    @pyd.model_validator(mode="after")
    def check_some_nonnull_inputs(self):
        if self.table_path is None and self.file_path is None:
            raise ValueError("table_path and/or file_path must be non-null")
        return self


class ModelConfig(pyd.BaseModel):
    experiment_id: str
    run_id: str
    framework: t.Optional[t.Literal["sklearn", "xgboost", "lightgbm"]] = None

    @pyd.computed_field  # type: ignore[misc]
    @property
    def mlflow_model_uri(self) -> str:
        return f"runs:/{self.run_id}/model"


class PreprocessingConfig(pyd.BaseModel):
    features: "FeaturesConfig"
    selection: "SelectionConfig"
    checkpoint: "CheckpointNthConfig | CheckpointFirstConfig | CheckpointLastConfig | CheckpointFirstAtNumCreditsEarnedConfig | CheckpointFirstWithinCohortConfig | CheckpointLastInEnrollmentYearConfig"
    target: "TargetGraduationConfig | TargetRetentionConfig | TargetCreditsEarnedConfig"
    splits: dict[t.Literal["train", "test", "validate"], float] = pyd.Field(
        default={"train": 0.6, "test": 0.2, "validate": 0.2},
        description=(
            "Mapping of name to fraction of the full datset belonging to a given 'split', "
            "which is a randomized subset used for different parts of the modeling process"
        ),
    )
    sample_class_weight: t.Optional[t.Literal["balanced"] | dict[object, int]] = (
        pyd.Field(
            default=None,
            description=(
                "Weights associated with classes in the form ``{class_label: weight}`` "
                "or 'balanced' to automatically adjust weights inversely proportional "
                "to class frequencies in the input data. "
                "If null (default), then sample weights are not computed."
            ),
        )
    )

    @pyd.field_validator("splits", mode="after")
    @classmethod
    def check_split_fractions(cls, value: dict) -> dict:
        if (sum_fracs := sum(value.values())) != 1.0:
            raise pyd.ValidationError(
                f"split fractions must sum up to 1.0, but input sums up to {sum_fracs}"
            )
        return value


class FeaturesConfig(pyd.BaseModel):
    min_passing_grade: float = pyd.Field(
        default=constants.DEFAULT_MIN_PASSING_GRADE,
        description="Minimum numeric grade considered by institution as 'passing'",
        gt=0.0,
        lt=4.0,
    )
    min_num_credits_full_time: float = pyd.Field(
        default=constants.DEFAULT_MIN_NUM_CREDITS_FULL_TIME,
        description=(
            "Minimum number of credits *attempted* per term for a student's "
            "enrollment intensity to be considered 'full-time'."
        ),
        gt=0.0,
        lt=20.0,
    )
    course_level_pattern: str = pyd.Field(
        default=constants.DEFAULT_COURSE_LEVEL_PATTERN,
        description=(
            "Regular expression patttern that extracts a course's 'level' "
            "from a PDP course_number field"
        ),
    )
    core_terms: set[str] = pyd.Field(
        default=constants.DEFAULT_CORE_TERMS,
        description=(
            "Set of terms that together comprise the 'core' of the academic year, "
            "in contrast with additional, usually shorter terms that may take place "
            "between core terms"
        ),
    )
    peak_covid_terms: set[tuple[str, str]] = pyd.Field(
        default=constants.DEFAULT_PEAK_COVID_TERMS,
        description=(
            "Set of (academic year, academic term) pairs considered by institution "
            "as 'peak' COVID, for use in control variables to account for pandemic effects"
        ),
    )
    key_course_subject_areas: t.Optional[t.List[t.Union[str, t.List[str]]]] = pyd.Field(
        default=None,
        description=(
            "One or more course subject areas (formatted as 2-digit CIP codes) "
            "for which custom features should be computed, can be a list or include a nested list"
            "Example: ['51', ['27', '48']], so you would get features for 51 alone and features for 27 and 48 combined."
        ),
    )
    key_course_ids: t.Optional[t.List[t.Union[str, t.List[str]]]] = pyd.Field(
        default=None,
        description=(
            "One or more course ids (formatted as '[COURSE_PREFIX][COURSE_NUMBER]') "
            "for which custom features should be computed, can be a list or include nested lists"
        ),
    )


class SelectionConfig(pyd.BaseModel):
    student_criteria: dict[str, object] = pyd.Field(
        default_factory=dict,
        description=(
            "Column name in modeling dataset mapped to one or more values that it must equal "
            "in order for the corresponding student to be considered 'eligible'. "
            "Multiple criteria are combined with a logical 'AND'."
        ),
    )
    intensity_time_limits: t.Optional[types.IntensityTimeLimitsType] = pyd.Field(
        default=None,
        description=(
            "Mapping of enrollment intensity value (e.g. 'FULL-TIME') to the max number "
            "years or terms (e.g. [4.0, 'year'], [12.0, 'term']) considered to be 'on-time' "
            "for a given target (e.g. graduation, credits earned), "
            "where the numeric values are for the time between 'checkpoint' and 'target' "
            "terms. Passing special '*' as the only key applies the corresponding time limits "
            "to all students, regardless of intensity."
        ),
    )
    params: dict[str, object] = pyd.Field(
        default_factory=dict,
        description="Any additional parameters needed to configure student selection",
    )


class CheckpointBaseConfig(pyd.BaseModel):
    name: str = pyd.Field(
        default=...,
        description="Descriptive name for checkpoint, used as a component in model name",
    )
    type_: types.CheckpointTypeType = pyd.Field(
        default=..., description="Type of checkpoint to which config is applied"
    )
    sort_cols: str | list[str] = pyd.Field(
        default="term_rank",
        description="Column(s) used to sort students' terms, typically chronologically.",
    )
    include_cols: t.Optional[list[str]] = pyd.Field(
        default=None,
        description="Optional subset of columns to include in checkpoint student-terms.",
    )


class CheckpointNthConfig(CheckpointBaseConfig):
    type_: types.CheckpointTypeType = "nth"
    n: int = pyd.Field(default=...)
    term_is_pre_cohort_col: t.Optional[str] = pyd.Field(default="term_is_pre_cohort")
    exclude_pre_cohort_terms: t.Optional[bool] = pyd.Field(default=True)
    term_is_core_col: t.Optional[str] = pyd.Field(default="term_is_core")
    exclude_non_core_terms: t.Optional[bool] = pyd.Field(default=True)
    enrollment_year_col: t.Optional[str] = pyd.Field(default=None)
    valid_enrollment_year: t.Optional[int] = pyd.Field(default=None)


class CheckpointFirstConfig(CheckpointBaseConfig):
    type_: types.CheckpointTypeType = "first"


class CheckpointLastConfig(CheckpointBaseConfig):
    type_: types.CheckpointTypeType = "last"


class CheckpointFirstAtNumCreditsEarnedConfig(CheckpointBaseConfig):
    type_: types.CheckpointTypeType = "first_at_num_credits_earned"
    min_num_credits: float = pyd.Field(default=...)
    num_credits_col: str = pyd.Field(default="num_credits_earned_cumsum")


class CheckpointFirstWithinCohortConfig(CheckpointBaseConfig):
    type_: types.CheckpointTypeType = "first_within_cohort"
    term_is_pre_cohort_col: str = pyd.Field(default="term_is_pre_cohort")


class CheckpointLastInEnrollmentYearConfig(CheckpointBaseConfig):
    type_: types.CheckpointTypeType = "last_in_enrollment_year"
    enrollment_year: float = pyd.Field(default=...)
    enrollment_year_col: str = pyd.Field(default="year_of_enrollment_at_cohort_inst")


class TargetBaseConfig(pyd.BaseModel):
    name: str = pyd.Field(
        default=...,
        description="Descriptive name for target, used as a component in model name",
    )
    type_: types.TargetTypeType = pyd.Field(
        default=..., description="Type of target to which config is applied"
    )


class TargetGraduationConfig(TargetBaseConfig):
    type_: types.TargetTypeType = "graduation"
    intensity_time_limits: types.IntensityTimeLimitsType
    years_to_degree_col: str
    num_terms_in_year: int = pyd.Field(default=4)
    max_term_rank: int | t.Literal["infer"]


class TargetRetentionConfig(TargetBaseConfig):
    type_: types.TargetTypeType = "retention"
    max_academic_year: str | t.Literal["infer"]


class TargetCreditsEarnedConfig(TargetBaseConfig):
    type_: types.TargetTypeType = "credits_earned"
    min_num_credits: float
    # TODO: is there any way to represent checkpoint arg in toml, given its dtype?
    intensity_time_limits: types.IntensityTimeLimitsType
    num_terms_in_year: int = pyd.Field(default=4)
    max_term_rank: int | t.Literal["infer"]


class ModelingConfig(pyd.BaseModel):
    feature_selection: t.Optional["FeatureSelectionConfig"] = None
    training: "TrainingConfig"
    evaluation: t.Optional["EvaluationConfig"] = None


class FeatureSelectionConfig(pyd.BaseModel):
    """
    See Also:
        - :func:`modeling.feature_selection.select_features()`
    """

    force_include_cols: t.Optional[list[str]] = None
    incomplete_threshold: float = 0.5
    low_variance_threshold: float = 0.0
    collinear_threshold: t.Optional[float] = 10.0


class TrainingConfig(pyd.BaseModel):
    """
    References:
        - https://docs.databricks.com/en/machine-learning/automl/automl-api-reference.html#classify
    """

    exclude_cols: t.Optional[list[str]] = pyd.Field(
        default=None,
        description="One or more column names in dataset to exclude from training.",
    )
    time_col: t.Optional[str] = pyd.Field(
        default=None,
        description=(
            "Column name in dataset used to split train/test/validate sets chronologically, "
            "as an alternative to the randomized assignment in ``split_col`` ."
        ),
    )
    exclude_frameworks: t.Optional[list[str]] = pyd.Field(
        default=None,
        description="List of algorithm frameworks that AutoML excludes from training.",
    )
    primary_metric: str = pyd.Field(
        default="log_loss",
        description="Metric used to evaluate and rank model performance.",
    )
    timeout_minutes: t.Optional[int] = pyd.Field(
        default=None,
        description="Maximum time to wait for AutoML trials to complete.",
    )


class EvaluationConfig(pyd.BaseModel):
    topn_runs_included: int = pyd.Field(
        default=3,
        description="Number of top-scoring mlflow runs to include for detailed evaluation",
    )


class InferenceConfig(pyd.BaseModel):
    num_top_features: int = pyd.Field(default=5)
    min_prob_pos_label: t.Optional[float] = 0.5
    # TODO: extend this configuration, maybe?



import functools as ft
import logging
import typing as t
from collections.abc import Collection

import pandas as pd

from ....utils import types
from . import constants, shared

LOGGER = logging.getLogger(__name__)


def add_features(
    df: pd.DataFrame,
    *,
    first_term_of_year: types.TermType = constants.DEFAULT_FIRST_TERM_OF_YEAR,  # type: ignore
    core_terms: set[types.TermType] = constants.DEFAULT_CORE_TERMS,  # type: ignore
    peak_covid_terms: set[tuple[str, str]] = constants.DEFAULT_PEAK_COVID_TERMS,
    year_col: str = "academic_year",
    term_col: str = "academic_term",
) -> pd.DataFrame:
    """
    Compute term-level features from pdp course dataset,
    and add as columns to ``df`` .

    Args:
        df
        first_term_of_year
        core_terms: Set of terms that together comprise the "core" of the academic year,
            in contrast with additional, usually shorter terms that may take place
            between core terms. Default value is {"FALL", "SPRING"}, which typically
            corresponds to a semester system; for schools on a trimester calendary,
            {"FALL", "WINTER", "SPRING"} is probably what you want.
        peak_covid_terms: Set of (year, term) pairs considered by the institution as
            occurring during "peak" COVID; for example, ``("2020-21", "SPRING")`` .
    """
    LOGGER.info("adding term features ...")
    noncore_terms: set[types.TermType] = set(df[term_col].unique()) - set(core_terms)
    df_term = (
        _get_unique_sorted_terms_df(df, year_col=year_col, term_col=term_col)
        # only need to compute features on unique terms, rather than at course-level
        # merging back into `df` afterwards ensures all rows have correct values
        .assign(
            term_id=ft.partial(shared.year_term, year_col=year_col, term_col=term_col),
            term_start_dt=ft.partial(
                shared.year_term_dt,
                col="term_id",
                bound="start",
                first_term_of_year=first_term_of_year,
            ),
            term_rank=ft.partial(term_rank, year_col=year_col, term_col=term_col),
            term_rank_core=ft.partial(
                term_rank,
                year_col=year_col,
                term_col=term_col,
                terms_subset=core_terms,
            ),
            term_rank_noncore=ft.partial(
                term_rank,
                year_col=year_col,
                term_col=term_col,
                terms_subset=noncore_terms,
            ),
            term_in_peak_covid=ft.partial(
                term_in_peak_covid,
                year_col=year_col,
                term_col=term_col,
                peak_covid_terms=peak_covid_terms,
            ),
            # yes, this is silly, but it helps a tricky feature computation later on
            term_is_core=ft.partial(
                term_in_subset, terms_subset=core_terms, term_col=term_col
            ),
            term_is_noncore=ft.partial(
                term_in_subset, terms_subset=noncore_terms, term_col=term_col
            ),
        )
    )
    return pd.merge(df, df_term, on=[year_col, term_col], how="inner")


def term_rank(
    df: pd.DataFrame,
    *,
    year_col: str = "academic_year",
    term_col: str = "academic_term",
    terms_subset: t.Optional[Collection[str]] = None,
) -> pd.Series:
    df_terms = (
        _get_unique_sorted_terms_df(df, year_col=year_col, term_col=term_col)
        if terms_subset is None
        else _get_unique_sorted_terms_df(
            df.loc[df[term_col].isin(terms_subset), :],
            year_col=year_col,
            term_col=term_col,
        )
    )
    df_terms_ranked = df_terms.assign(
        term_rank=lambda df: pd.Series(list(range(len(df))))
    )
    # left-join back into df, so this works if df rows are at the course *or* term level
    term_id_cols = [year_col, term_col]
    return (
        pd.merge(df[term_id_cols], df_terms_ranked, on=term_id_cols, how="left")
        .loc[:, "term_rank"]
        .rename(None)
        .astype("Int8")
    )


def term_in_peak_covid(
    df: pd.DataFrame,
    *,
    year_col: str = "academic_year",
    term_col: str = "academic_term",
    peak_covid_terms: set[tuple[str, str]],
) -> pd.Series:
    return pd.Series(
        pd.MultiIndex.from_frame(df[[year_col, term_col]]).isin(peak_covid_terms)
    )


def term_in_subset(
    df: pd.DataFrame, terms_subset: set[types.TermType], term_col: str = "academic_term"
) -> pd.Series:
    return df[term_col].isin(terms_subset).astype("boolean")


def _get_unique_sorted_terms_df(
    df: pd.DataFrame,
    *,
    year_col: str = "academic_year",
    term_col: str = "academic_term",
) -> pd.DataFrame:
    return (
        df[[year_col, term_col]]
        # dedupe more convenient op than groupby([year_col, term_col])
        .drop_duplicates(ignore_index=True)
        # null year and/or term values aren't relevant/useful here
        .dropna(axis="index", how="any")
        # assumes year col is alphanumerically sortable, term col is categorically ordered
        .sort_values(by=[year_col, term_col], ignore_index=True)
    )



import pytest
from unittest.mock import MagicMock

from student_success_tool.reporting.sections.registry import SectionRegistry
from student_success_tool.reporting.sections.attribute_sections import (
    register_attribute_sections,
)
from student_success_tool.reporting.utils.formatting import Formatting


@pytest.fixture
def mock_card():
    card = MagicMock()
    card.format = Formatting()
    return card


def test_outcome_graduation_and_full_time_limit(mock_card):
    mock_card.cfg.preprocessing.target.name = "graduation"
    mock_card.cfg.preprocessing.selection.intensity_time_limits = {
        "FULL-TIME": (2.0, "year"),
        "PART-TIME": (3, "year"),
    }

    registry = SectionRegistry()
    register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert (
        "graduation within 2 years for full-time students"
        in rendered["outcome_section"]
    )
    assert "within 3 years for part-time students" in rendered["outcome_section"]


def test_outcome_missing_target_or_limits(mock_card):
    mock_card.cfg.preprocessing.target.name = ""
    mock_card.cfg.preprocessing.selection.intensity_time_limits = {}

    registry = SectionRegistry()
    register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert (
        rendered["outcome_section"] == "**Target or Time Limit Information Not Found**"
    )


def test_target_population_section(mock_card):
    mock_card.cfg.preprocessing.selection.student_criteria = {
        "degree": ["bachelor's", "associate's"],
        "status": "full-time",
    }

    registry = SectionRegistry()
    register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert "Degree" in rendered["target_population_section"]
    assert "- Bachelor's" in rendered["target_population_section"]
    assert "- Associate's" in rendered["target_population_section"]
    assert "Status" in rendered["target_population_section"]
    assert "- Full-Time" in rendered["target_population_section"]


def test_checkpoint_section_with_credit(mock_card):
    mock_card.cfg.preprocessing.checkpoint.name = "credit_45"
    mock_card.cfg.preprocessing.checkpoint.params = {"min_num_credits": 45}

    registry = SectionRegistry()
    register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert "45 credits" in rendered["checkpoint_section"]


def test_checkpoint_section_with_semester(mock_card):
    mock_card.cfg.preprocessing.checkpoint.name = "first_semester"
    mock_card.cfg.preprocessing.checkpoint.params = {}

    registry = SectionRegistry()
    register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert "first semester" in rendered["checkpoint_section"]


def test_checkpoint_section_unknown(mock_card):
    mock_card.cfg.preprocessing.checkpoint.name = "unknown_metric"
    mock_card.cfg.preprocessing.checkpoint.params = {}

    registry = SectionRegistry()
    register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert rendered["checkpoint_section"] == "**Checkpoint Information Not Found**"



DEFAULT_MIN_PASSING_GRADE = 1.0
DEFAULT_MIN_NUM_CREDITS_FULL_TIME = 12.0
DEFAULT_COURSE_CREDIT_CHECK = 12
DEFAULT_FIRST_TERM_OF_YEAR = "FALL"
DEFAULT_CORE_TERMS = {"FALL", "SPRING"}
DEFAULT_COURSE_LEVEL_PATTERN = r"^(?P<course_level>\d)\d{2}(?:[A-Z]{,2})?$"
DEFAULT_PEAK_COVID_TERMS = {
    ("2019-20", "SPRING"),  # Spring 2020
    ("2019-20", "SUMMER"),  # Summer 2020
    ("2020-21", "FALL"),  # Fall 2020
    ("2020-21", "WINTER"),  # Winter 2020/2021
    ("2020-21", "SPRING"),  # Spring 2021
    ("2020-21", "SUMMER"),  # Summer 2021
}

NUM_COURSE_FEATURE_COL_PREFIX = "num_courses"
FRAC_COURSE_FEATURE_COL_PREFIX = "frac_courses"
DUMMY_COURSE_FEATURE_COL_PREFIX = "took"



import functools as ft
import logging

import pandas as pd

from . import constants, shared

LOGGER = logging.getLogger(__name__)

NON_NUMERIC_GRADES = {"A", "F", "I", "M", "O", "P", "W"}
NON_PASS_FAIL_GRADES = {"A", "I", "M", "O", "W"}
NON_COMPLETE_GRADES = {"I", "W"}


def add_features(
    df: pd.DataFrame,
    *,
    min_passing_grade: float = constants.DEFAULT_MIN_PASSING_GRADE,
    course_level_pattern: str = constants.DEFAULT_COURSE_LEVEL_PATTERN,
) -> pd.DataFrame:
    """
    Compute course-level features from a pdp course dataset,
    and add as columns to ``df`` .

    Args:
        df
        min_passing_grade: Minimum numeric grade considered by institution as "passing".
            Note that this is represented as a float, while grades are strings
            since the values include both numeric and alpha-categorical values.
            This value is only compared against numeric grades; relevant categoricals
            are handled appropriately, e.g. "P" => "Pass" is always considered "passing".
        course_level_pattern: Regex string that extracts a course's level from its number
            (e.g. 1 from "101"). *Must* include exactly one capture group,
            which is taken to be the course level.
    """
    LOGGER.info("adding course features ...")
    return df.assign(
        course_id=course_id,
        course_subject_area=course_subject_area,
        course_passed=ft.partial(course_passed, min_passing_grade=min_passing_grade),
        course_completed=course_completed,
        course_level=ft.partial(course_level, pattern=course_level_pattern),
        course_grade_numeric=course_grade_numeric,
        course_grade=course_grade,
    )


def course_id(
    df: pd.DataFrame,
    *,
    prefix_col: str = "course_prefix",
    number_col: str = "course_number",
) -> pd.Series:
    return df[prefix_col].str.cat(df[number_col], sep="")


def course_subject_area(df: pd.DataFrame, *, col: str = "course_cip") -> pd.Series:
    return shared.extract_short_cip_code(df[col])


def course_passed(
    df: pd.DataFrame, *, col: str = "grade", min_passing_grade: float
) -> pd.Series:
    series = (
        df[col]
        .astype("string")
        .map(
            ft.partial(_grade_is_passing, min_passing_grade=min_passing_grade),  # type: ignore
            na_action="ignore",
        )
        .astype("boolean")
    )
    assert isinstance(series, pd.Series)  # type guard
    return series


def course_completed(df: pd.DataFrame, *, col: str = "grade") -> pd.Series:
    return ~(df[col].astype("string").isin(NON_COMPLETE_GRADES))


def course_level(
    df: pd.DataFrame, *, col: str = "course_number", pattern: str
) -> pd.Series:
    return (
        df[col]
        .astype("string")
        .str.strip()
        .str.extract(pattern, expand=False)
        .astype("Int8")
    )


def course_grade_numeric(df: pd.DataFrame, *, col: str = "grade") -> pd.Series:
    return df[col].mask(df[col].isin(NON_NUMERIC_GRADES), pd.NA).astype("Float32")


def course_grade(
    df: pd.DataFrame,
    *,
    grade_col: str = "grade",
    grade_num_col: str = "course_grade_numeric",
) -> pd.Series:
    non_numeric_grades = (
        df[grade_col]
        .mask(~df[grade_col].isin(NON_NUMERIC_GRADES), pd.NA)
        # frustratingly, pdp uses "A" grade to indicate "Audit", which is just begging
        # for confusion with the usual meaning of an "A" grade :/
        # let's replace it with "AUDIT" for clarity, and so we can safely combine
        # non-numeric grades with derived letter grades below
        .replace("A", value="AUDIT")
        # similarly, "O" looks like "0", so let's replace with "OTHER" for clarity
        .replace("O", value="OTHER")
        .astype("string")
    )
    letter_grades = pd.cut(
        df[grade_num_col],
        # pandas' binning args here are bad if you want (standard!) left-inclusive bins
        # and *labels* for those bins; despite appearances, binning is like so:
        # [0, 0.7) => F, [0.7, 1.7) => D, [1.7, 2.7) => C, [2.7, 3.7) => B, [3.7, 4.0] => A
        bins=[0.0, 0.69, 1.69, 2.69, 3.69, 4.01],
        labels=["F", "D", "C", "B", "A"],
        right=True,
        include_lowest=True,
    ).astype("string")
    # NOTE: this assumes that "F" ("Fail") grades are equivalent to "F" letter grades
    return non_numeric_grades.combine_first(letter_grades)


def _grade_is_passing(grade: str, min_passing_grade: float) -> bool | None:
    if grade in NON_PASS_FAIL_GRADES:
        return None
    elif grade == "P":
        return True
    elif grade == "F":
        return False
    else:
        return float(grade) >= min_passing_grade



"""
HACK! This shows a concrete example of data preprocessing, encapsulated within a function.
Typically, this logic is implemented piecewise in the context of a Databricks notebook
(see: 01-preprocess-data-TEMPLATE.py).
"""

import functools as ft
import typing as t

import pandas as pd

from student_success_tool import modeling, schemas, targets
from student_success_tool.preprocessing.pdp import dataops


def preprocess_data(
    df_course: pd.DataFrame,
    df_cohort: pd.DataFrame,
    *,
    run_type: t.Literal["predict", "train"],
    student_id_col: str = "student_guid",
    cfg: schemas.pdp.PDPProjectConfigV2,
) -> pd.DataFrame:
    """
    Args:
        df_course
        df_cohort
        run_type
        student_id_col
        cfg
    """
    df_student_terms = dataops.make_student_term_dataset(
        df_cohort, df_course, **cfg.preprocessing.features.model_dump()
    )
    if run_type == "train":
        df_modeling = targets.pdp.failure_to_retain.make_labeled_dataset(
            df_student_terms, **cfg.preprocessing.target.params
        )
    else:
        eligible_students = targets.pdp.shared.select_students_by_criteria(
            df_student_terms,
            student_id_cols=student_id_col,
            **cfg.preprocessing.target.params["student_criteria"],
        )
        max_term_rank = df_student_terms["term_rank"].max()
        df_modeling = pd.merge(
            df_student_terms.loc[df_student_terms["term_rank"].eq(max_term_rank), :],
            eligible_students,
            on=student_id_col,
            how="inner",
        )
    df_modeling = dataops.clean_up_labeled_dataset_cols_and_vals(df_modeling)
    if run_type == "train":
        if cfg.split_col is not None and cfg.preprocessing.splits:
            df_modeling = df_modeling.assign(
                **{
                    cfg.split_col: ft.partial(
                        modeling.utils.compute_dataset_splits, seed=cfg.random_state
                    )
                }
            )
        if cfg.sample_weight_col is not None and cfg.preprocessing.sample_class_weight:
            df_modeling = df_modeling.assign(
                **{
                    cfg.sample_weight_col: ft.partial(
                        modeling.utils.compute_sample_weights,
                        target_col=cfg.target_col,
                        class_weight=cfg.preprocessing.sample_class_weight,
                    )
                }
            )
    return df_modeling



import matplotlib.pyplot as plt
import mlflow.tracking
import numpy as np
import pandas as pd
import pytest

from student_success_tool.modeling import bias_detection

np.random.seed(42)

# HACK: log artifacts to tmp dir so they don't clutter repo
mlflow.set_tracking_uri("file:///tmp/mlflow")


@pytest.fixture
def patch_mlflow(monkeypatch):
    """Patch the mlflow.search_runs function."""

    def mock_search_runs():
        # This mock function does nothing and will be overridden by each test case
        pass

    monkeypatch.setattr(mlflow.tracking.MlflowClient, "search_runs", mock_search_runs)


@pytest.fixture
def mock_mlflow(monkeypatch):
    monkeypatch.setattr(mlflow, "log_figure", lambda *args, **kwargs: None)
    monkeypatch.setattr(
        bias_detection, "log_group_metrics_to_mlflow", lambda *args, **kwargs: None
    )
    monkeypatch.setattr(
        bias_detection, "log_subgroup_metrics_to_mlflow", lambda *args, **kwargs: None
    )
    monkeypatch.setattr(
        bias_detection, "log_bias_flags_to_mlflow", lambda *args, **kwargs: None
    )


@pytest.fixture
def mock_helpers(monkeypatch):
    monkeypatch.setattr(bias_detection, "flag_bias", lambda fnr_data: fnr_data)
    monkeypatch.setattr(bias_detection, "plot_fnr_group", lambda fnr_data: plt.figure())
    monkeypatch.setattr(
        bias_detection,
        "calculate_fnr_and_ci",
        lambda y_true, y_pred: (0.5, 0.4, 0.6, sum(y_true)),
    )


def test_compute_group_bias_metrics(mock_helpers):
    df = pd.DataFrame(
        {
            "group_col": ["A", "A", "B", "B"],
            "target": [1, 0, 1, 0],
            "pred": [1, 0, 0, 1],
            "prob": [0.9, 0.2, 0.3, 0.8],
            "sample_weight_col": [1.2, 0.8, 1.2, 0.8],
        }
    )

    bias_metrics, perf_metrics, fnr_data = bias_detection.compute_group_bias_metrics(
        split_data=df,
        split_name="test",
        group_col="group_col",
        target_col="target",
        pred_col="pred",
        pred_prob_col="prob",
        pos_label=1,
        sample_weight_col="sample_weight_col",
    )

    assert isinstance(bias_metrics, list)
    assert isinstance(perf_metrics, list)
    assert isinstance(fnr_data, list)
    assert all("FNR" in m for m in bias_metrics)
    assert all("subgroup" in f for f in fnr_data)
    assert all(f["split_name"] == "test" for f in fnr_data)
    assert all(f["fnr"] > 0 and f["ci"][0] < f["fnr"] < f["ci"][1] for f in fnr_data)


@pytest.mark.parametrize(
    "targets, preds, expected_fnr, expected_ci_lower, expected_ci_upper, valid_samples_flag",
    [
        (
            pd.Series(np.random.choice([False, True], size=500)),  # Use bool values
            pd.Series(np.random.choice([False, True], size=500)),
            0.504685,
            0.444092,
            0.565278,
            256,
        ),
        (
            pd.Series([True] * 500, dtype=bool),
            pd.Series([True] * 500, dtype=bool),
            0,
            0,
            0,
            500,
        ),
        (
            pd.Series([False] * 500, dtype=bool),
            pd.Series([False] * 500, dtype=bool),
            0,
            0,
            0,
            0,
        ),
    ],
)
def test_calculate_fnr_and_ci(
    targets,
    preds,
    expected_fnr,
    expected_ci_lower,
    expected_ci_upper,
    valid_samples_flag,
):
    fnr, ci_lower, ci_upper, valid_samples_flag = bias_detection.calculate_fnr_and_ci(
        targets, preds
    )
    assert np.isclose(fnr, expected_fnr, equal_nan=True)
    assert np.isclose(ci_lower, expected_ci_lower, equal_nan=True)
    assert np.isclose(ci_upper, expected_ci_upper, equal_nan=True)


@pytest.mark.parametrize(
    "ci1, ci2, expected",
    [
        ((0.1, 0.3), (0.2, 0.4), True),
        ((0.1, 0.2), (0.3, 0.4), False),
        ((0.05, 0.15), (0.1, 0.2), True),
    ],
)
def test_check_ci_overlap(ci1, ci2, expected):
    assert bias_detection.check_ci_overlap(ci1, ci2) == expected


@pytest.mark.parametrize(
    "fnr1, fnr2, denom1, denom2, expected_p",
    [
        (0.2, 0.25, 100, 100, 0.3963327),
        (0.1, 0.15, 20, 35, np.nan),
        (0.3, 0.1, 200, 200, 2.4e-07),
    ],
)
def test_z_test_fnr_difference(fnr1, fnr2, denom1, denom2, expected_p):
    p_value = bias_detection.z_test_fnr_difference(fnr1, fnr2, denom1, denom2)
    assert np.isclose(p_value, expected_p, equal_nan=True)


@pytest.mark.parametrize(
    "group, sub1, sub2, fnr_percentage_difference, bias_type, split_name, flag, p, expected",
    [
        (
            "Gender",
            "Male",
            "Female",
            0.12,
            "non-overlapping confidence intervals",
            "train",
            "🔴 HIGH BIAS",
            0.005,
            {
                "group": "Gender",
                "subgroups": "Male vs Female",
                "fnr_percentage_difference": 0.12,
                "type": "non-overlapping confidence intervals with a p-value of 0.005",
                "split_name": "train",
                "flag": "🔴 HIGH BIAS",
            },
        ),
    ],
)
def test_generate_bias_flag(
    group,
    sub1,
    sub2,
    fnr_percentage_difference,
    bias_type,
    split_name,
    flag,
    p,
    expected,
):
    assert (
        bias_detection.generate_bias_flag(
            group,
            sub1,
            sub2,
            fnr_percentage_difference,
            bias_type,
            split_name,
            flag,
            p,
        )
        == expected
    )


@pytest.mark.parametrize(
    "fnr_data, expected_sg1, expected_sg2, expected_flag",
    [
        # Case where the ordering should be respected (greater FNR first)
        (
            [
                {
                    "group": "Race",
                    "subgroup": "GroupA",
                    "fnr": 0.30,
                    "size": 100,
                    "ci": (0.25, 0.35),
                    "number_of_positive_samples": 20,
                    "split_name": "validation",
                },
                {
                    "group": "Race",
                    "subgroup": "GroupB",
                    "fnr": 0.10,
                    "size": 100,
                    "ci": (0.05, 0.15),
                    "number_of_positive_samples": 20,
                    "split_name": "validation",
                },
            ],
            "GroupA",
            "GroupB",
            "🔴 HIGH BIAS",
        ),
    ],
)
def test_flag_bias(fnr_data, expected_sg1, expected_sg2, expected_flag):
    def mock_z_test_fnr_difference(fnr1, fnr2, n1, n2):
        return 0.005  # Significant p-value to trigger flagging

    def mock_check_ci_overlap(ci1, ci2):
        return False  # Non-overlapping

    def mock_generate_bias_flag(
        group, sg1, sg2, fnr_diff, reason, split_name, flag, p_value
    ):
        return {
            "group": group,
            "sg1": sg1,
            "sg2": sg2,
            "flag": flag,
            "fnr_diff": fnr_diff,
            "p_value": p_value,
        }

    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr(
        bias_detection, "z_test_fnr_difference", mock_z_test_fnr_difference
    )
    monkeypatch.setattr(bias_detection, "check_ci_overlap", mock_check_ci_overlap)
    monkeypatch.setattr(bias_detection, "generate_bias_flag", mock_generate_bias_flag)

    result = bias_detection.flag_bias(
        fnr_data,
        high_bias_thresh=0.15,
        moderate_bias_thresh=0.1,
        low_bias_thresh=0.05,
        min_sample_ratio=0.1,
    )

    assert len(result) == 1
    flag = result[0]
    assert flag["sg1"] == expected_sg1
    assert flag["sg2"] == expected_sg2
    assert flag["flag"] == expected_flag

    monkeypatch.undo()



from .metric_sections import register_metric_sections
from .attribute_sections import register_attribute_sections
from .data_sections import register_data_sections
from .evaluation_sections import register_evaluation_sections
from .bias_sections import register_bias_sections


def register_sections(card, registry):
    register_metric_sections(card, registry)
    register_attribute_sections(card, registry)
    register_data_sections(card, registry)
    register_evaluation_sections(card, registry)
    register_bias_sections(card, registry)



import datetime
import functools as ft
import logging
import re
import typing as t

import pandas as pd

from .. import utils
from . import features
from .features.pdp import constants

LOGGER = logging.getLogger(__name__)


def make_student_term_dataset(
    df_cohort: pd.DataFrame,
    df_course: pd.DataFrame,
    *,
    min_passing_grade: float = constants.DEFAULT_MIN_PASSING_GRADE,
    min_num_credits_full_time: float = constants.DEFAULT_MIN_NUM_CREDITS_FULL_TIME,
    course_level_pattern: str = constants.DEFAULT_COURSE_LEVEL_PATTERN,
    core_terms: set[str] = constants.DEFAULT_CORE_TERMS,
    peak_covid_terms: set[tuple[str, str]] = constants.DEFAULT_PEAK_COVID_TERMS,
    key_course_subject_areas: t.Optional[list[str]] = None,
    key_course_ids: t.Optional[list[str]] = None,
) -> pd.DataFrame:
    """
    Make a student-term dataset from raw cohort- and course-level datasets,
    including many features generated at the student-, course-, term-, section-,
    and student-term levels, as well as cumulatively over student-terms.

    Args:
        df_cohort: As output by :func:`dataio.read_raw_pdp_cohort_data_from_file()` .
        df_course: As output by :func:`dataio.read_raw_pdp_course_data_from_file()` .
        min_passing_grade: Minimum numeric grade considered by institution as "passing".
            Default value is 1.0, i.e. a "D" grade or better.
        min_num_credits_full_time: Minimum number of credits *attempted* per term
            for a student's enrollment intensity to be considered "full-time".
            Default value is 12.0.
        core_terms: Set of terms that together comprise the "core" of the academic year,
            in contrast with additional, usually shorter terms that may take place
            between core terms. Default value is {"FALL", "SPRING"}, which typically
            corresponds to a semester system; for schools on a trimester calendary,
            {"FALL", "WINTER", "SPRING"} is probably what you want.
        course_level_pattern
        peak_covid_terms
        key_course_subject_areas
        key_courses_ids

    References:
        - https://bigfuture.collegeboard.org/plan-for-college/get-started/how-to-convert-gpa-4.0-scale
    """
    first_term_of_year = infer_first_term_of_year(df_course["academic_term"])
    df_students = (
        df_cohort.pipe(standardize_cohort_dataset)
        .pipe(features.pdp.student.add_features, first_term_of_year=first_term_of_year)
    )  # fmt: skip
    df_courses_plus = (
        df_course.pipe(standardize_course_dataset)
        .pipe(
            features.pdp.course.add_features,
            min_passing_grade=min_passing_grade,
            course_level_pattern=course_level_pattern,
        )
        .pipe(
            features.pdp.term.add_features,
            first_term_of_year=first_term_of_year,
            core_terms=core_terms,  # type: ignore
            peak_covid_terms=peak_covid_terms,
        )
        .pipe(
            features.pdp.section.add_features,
            section_id_cols=["term_id", "course_id", "section_id"],
        )
    )
    df_student_terms = (
        features.pdp.student_term.aggregate_from_course_level_features(
            df_courses_plus,
            student_term_id_cols=["student_id", "term_id"],
            min_passing_grade=min_passing_grade,
            key_course_subject_areas=key_course_subject_areas,
            key_course_ids=key_course_ids,
        )
        .merge(df_students, how="inner", on=["institution_id", "student_id"])
        .pipe(
            features.pdp.student_term.add_features,
            min_num_credits_full_time=min_num_credits_full_time,
        )
    )
    df_student_terms_plus = (
        features.pdp.cumulative.add_features(
            df_student_terms,
            student_id_cols=["institution_id", "student_id"],
            sort_cols=["academic_year", "academic_term"],
        )
        # NOTE: it's important to standardize column names here to avoid name mismatches
        # when features are generated here (on-the-fly) as opposed to read (pre-computed)
        # from a delta table; spark can be configured to behave nicely...
        # but let's not take any chances
        .rename(columns=utils.misc.convert_to_snake_case)
    )
    return df_student_terms_plus


def standardize_cohort_dataset(df: pd.DataFrame) -> pd.DataFrame:
    """
    Drop some columns from raw cohort dataset.

    Args:
        df: As output by :func:`dataio.read_raw_pdp_cohort_data_from_file()` .
    """
    LOGGER.info("standardizing cohort dataset ...")
    df_trf = (
        # drop redundant/useless/unwanted cols
        df.pipe(
            drop_columns_safely,
            cols=[
                # not a viable target variable, but highly correlated with it
                "time_to_credential",
                # not all demographics used for target variable bias checks
                "incarcerated_status",
                "military_status",
                "employment_status",
                "disability_status",
                "naspa_first_generation",
                # redundant
                "attendance_status_term_1",
                # covered indirectly by course dataset fields/features
                "gateway_math_status",
                "gateway_english_status",
                "attempted_gateway_math_year_1",
                "attempted_gateway_english_year_1",
                "completed_gateway_math_year_1",
                "completed_gateway_english_year_1",
                "gateway_math_grade_y_1",
                "gateway_english_grade_y_1",
                "attempted_dev_math_y_1",
                "attempted_dev_english_y_1",
                "completed_dev_math_y_1",
                "completed_dev_english_y_1",
                # let's assume we don't need other institution "demographics"
                "most_recent_bachelors_at_other_institution_state",
                "most_recent_associates_or_certificate_at_other_institution_state",
                "most_recent_last_enrollment_at_other_institution_state",
                "first_bachelors_at_other_institution_state",
                "first_associates_or_certificate_at_other_institution_state",
                "most_recent_bachelors_at_other_institution_carnegie",
                "most_recent_associates_or_certificate_at_other_institution_carnegie",
                "most_recent_last_enrollment_at_other_institution_carnegie",
                "first_bachelors_at_other_institution_carnegie",
                "first_associates_or_certificate_at_other_institution_carnegie",
                "most_recent_bachelors_at_other_institution_locale",
                "most_recent_associates_or_certificate_at_other_institution_locale",
                "most_recent_last_enrollment_at_other_institution_locale",
                "first_bachelors_at_other_institution_locale",
                "first_associates_or_certificate_at_other_institution_locale",
            ],
        )
        # as pdp adds more raw data columns, we'll want to ensure their presence here
        # so that feature generation code doesn't become a cascading mess of "if" checks
        .pipe(
            add_empty_cols_if_missing,
            col_val_dtypes={
                "years_to_latest_associates_at_cohort_inst": (None, "Int8"),
                "years_to_latest_certificate_at_cohort_inst": (None, "Int8"),
                "years_to_latest_associates_at_other_inst": (None, "Int8"),
                "years_to_latest_certificate_at_other_inst": (None, "Int8"),
                "first_year_to_associates_at_cohort_inst": (None, "Int8"),
                "first_year_to_certificate_at_cohort_inst": (None, "Int8"),
                "first_year_to_associates_at_other_inst": (None, "Int8"),
                "first_year_to_certificate_at_other_inst": (None, "Int8"),
            },
        )
    )
    return df_trf


def standardize_course_dataset(df: pd.DataFrame) -> pd.DataFrame:
    """
    Drop some columns and anomalous rows from raw course dataset.

    Args:
        df: As output by :func:`dataio.read_raw_pdp_course_data_from_file()` .
    """
    return (
        # drop rows for which we're missing key information
        df.pipe(drop_course_rows_missing_identifiers)
        # drop redundant/useless/unwanted cols
        .pipe(
            drop_columns_safely,
            cols=[
                # student demographics found in raw cohort dataset
                "cohort",
                "cohort_term",
                "student_age",
                "race",
                "ethnicity",
                "gender",
                # course name and aspects of core-ness not needed
                "course_name",
                "core_course_type",
                "core_competency_completed",
                "credential_engine_identifier",
                # enrollment record at other insts not needed
                "enrollment_record_at_other_institution_s_state_s",
                "enrollment_record_at_other_institution_s_carnegie_s",
                "enrollment_record_at_other_institution_s_locale_s",
            ],
        )
        # as pdp adds more raw data columns, we'll want to ensure their presence here
        # so that feature generation code doesn't become a cascading mess of "if" checks
        .pipe(
            add_empty_cols_if_missing,
            col_val_dtypes={"term_program_of_study": (None, "string")},
        )
    )


def add_empty_cols_if_missing(
    df: pd.DataFrame,
    *,
    col_val_dtypes: dict[
        str, tuple[t.Optional[str | bool | int | float | datetime.datetime], str]
    ],
) -> pd.DataFrame:
    """
    Add empty columns to ``df`` with names given by keys in ``col_val_dtypes``
    matched to values representing the null value and underlying dtype assigned to it
    -- provided the columns aren't already present in the dataframe.
    """
    return df.assign(
        **{
            col: pd.Series(data=val, index=df.index, dtype=dtype)
            for col, (val, dtype) in col_val_dtypes.items()
            if col not in df.columns
        }
    )


def clean_up_labeled_dataset_cols_and_vals(
    df: pd.DataFrame, num_credits_col: str = "cumsum_num_credits_earned"
) -> pd.DataFrame:
    """
    Drop a bunch of columns not needed or wanted for modeling, and set to null
    any values corresponding to time after a student's current year of enrollment.

    Args:
        df: DataFrame as created with features and targets and limited to the checkpoint term.
        num_credits_col: Name of the column containing cumulative earned credits.
    """
    num_credit_check = constants.DEFAULT_COURSE_CREDIT_CHECK
    credit_pattern = re.compile(rf"in_{num_credit_check}_creds")
    # To prevent data leakage, students that have not reached the 12 credits and not taken the course
    # by the checkpoint term (which this data is limited to at the time of this function),
    # will have the applicable in_12_credits columns set to null.
    for col in df.columns:
        if credit_pattern.search(col):
            df[col] = df[col].mask((df[num_credits_col] < num_credit_check))

    return (
        # drop many columns that *should not* be included as features in a model
        df.pipe(
            drop_columns_safely,
            cols=[
                # metadata
                "institution_id",
                "term_id",
                "academic_year",
                # "academic_term",  # keeping this to see if useful
                "cohort",
                # "cohort_term",  # keeping this to see if useful
                "cohort_id",
                "term_rank",
                "min_student_term_rank",
                "term_rank_core",
                "min_student_term_rank_core",
                "term_is_core",
                "term_rank_noncore",
                "min_student_term_rank_noncore",
                "term_is_noncore",
                # columns used to derive other features, but not features themselves
                # "grade",  # TODO: should this be course_grade?
                "course_ids",
                "course_subjects",
                "course_subject_areas",
                "min_student_term_rank",
                "min_student_term_rank_core",
                "min_student_term_rank_noncore",
                "sections_num_students_enrolled",
                "sections_num_students_passed",
                "sections_num_students_completed",
                "term_start_dt",
                "cohort_start_dt",
                "pell_status_first_year",
                # "outcome" variables / likely sources of data leakage
                "retention",
                "persistence",
                # years to bachelors
                "years_to_bachelors_at_cohort_inst",
                "years_to_bachelor_at_other_inst",
                "first_year_to_bachelors_at_cohort_inst",
                "first_year_to_bachelor_at_other_inst",
                # years to associates
                "years_to_latest_associates_at_cohort_inst",
                "years_to_latest_associates_at_other_inst",
                "first_year_to_associates_at_cohort_inst",
                "first_year_to_associates_at_other_inst",
                # years to associates / certificate
                "years_to_associates_or_certificate_at_cohort_inst",
                "years_to_associates_or_certificate_at_other_inst",
                "first_year_to_associates_or_certificate_at_cohort_inst",
                "first_year_to_associates_or_certificate_at_other_inst",
                # years of last enrollment
                "years_of_last_enrollment_at_cohort_institution",
                "years_of_last_enrollment_at_other_institution",
            ],
        ).assign(
            # keep "first year to X credential at Y inst"  or "years to latest X credential at Y inst" values if they occurred
            # in any year prior to the current year of enrollment; otherwise, set to null
            # in this case, the values themselves represent years
            **{
                col: ft.partial(_mask_year_values_based_on_enrollment_year, col=col)
                for col in df.columns[
                    df.columns.str.contains(
                        r"^(?:first_year_to_certificate|years_to_latest_certificate)"
                    )
                ].tolist()
            }
            # keep values in "*_year_X" columns if they occurred in any year prior
            # to the current year of enrollment; otherwise, set to null
            # in this case, the columns themselves represent years
            | {
                col: ft.partial(_mask_year_column_based_on_enrollment_year, col=col)
                for col in df.columns[df.columns.str.contains(r"_year_\d$")].tolist()
            }
            # do the same thing, only by term for term columns: "*_term_X"
            | {
                col: ft.partial(_mask_term_column_based_on_enrollment_term, col=col)
                for col in df.columns[df.columns.str.contains(r"_term_\d$")].tolist()
            }
        )
    )


def _mask_year_values_based_on_enrollment_year(
    df: pd.DataFrame,
    *,
    col: str,
    enrollment_year_col: str = "year_of_enrollment_at_cohort_inst",
) -> pd.Series:
    return df[col].mask(df[col].ge(df[enrollment_year_col]), other=pd.NA)


def _mask_year_column_based_on_enrollment_year(
    df: pd.DataFrame,
    *,
    col: str,
    enrollment_year_col: str = "year_of_enrollment_at_cohort_inst",
) -> pd.Series:
    if match := re.search(r"_year_(?P<yr>\d)$", col):
        col_year = int(match.groupdict()["yr"])
    else:
        raise ValueError(f"column '{col}' does not end with '_year_NUM'")
    return df[col].mask(df[enrollment_year_col].le(col_year), other=pd.NA)


def _mask_term_column_based_on_enrollment_term(
    df: pd.DataFrame,
    *,
    col: str,
    enrollment_term_col: str = "cumnum_terms_enrolled",
) -> pd.Series:
    if match := re.search(r"_term_(?P<num>\d)$", col):
        col_term = int(match.groupdict()["num"])
    else:
        raise ValueError(f"column '{col}' does not end with '_term_NUM'")
    return df[col].mask(df[enrollment_term_col].lt(col_term), other=pd.NA)


def drop_course_rows_missing_identifiers(df: pd.DataFrame) -> pd.DataFrame:
    """
    Drop rows from raw course dataset missing key course identifiers,
    specifically course prefix and number, which supposedly are partial records
    from students' enrollments at *other* institutions -- not wanted here!
    """
    num_rows_before = len(df)
    df = (
        df.loc[df[["course_prefix", "course_number"]].notna().all(axis=1), :]
        # reset range index so there are no gaps, which can cause subtle errors
        # when using `pd.DataFrame.assign()` to add features
        .reset_index(drop=True)
    )
    num_rows_after = len(df)
    if num_rows_after < num_rows_before:
        LOGGER.warning(
            "dropped %s rows from course dataset owing to missing identifiers",
            num_rows_before - num_rows_after,
        )
    return df


def drop_columns_safely(df: pd.DataFrame, *, cols: list[str]) -> pd.DataFrame:
    """
    Drop ``cols`` from ``df`` *safely*: If any are missing, log a clear warning,
    then drop the non-missing columns from the DataFrame without crashing.

    Args:
        df
        cols
    """
    df_cols = set(df.columns)
    drop_cols = set(cols) & df_cols
    if missing_cols := (set(cols) - df_cols):
        LOGGER.warning(
            "%s column%s not found in df: %s",
            len(missing_cols),
            "s" if len(missing_cols) > 1 else "",
            missing_cols,
        )
    df_trf = df.drop(columns=list(drop_cols))
    LOGGER.info("dropped %s columns safely", len(drop_cols))
    return df_trf


def infer_first_term_of_year(s: pd.Series) -> utils.types.TermType:
    """
    Infer the first term of the (academic) year by the ordering of its categorical values.

    See Also:
        - :class:`schemas.base.TermField()`
    """
    if isinstance(s.dtype, pd.CategoricalDtype) and s.cat.ordered is True:
        first_term_of_year = s.cat.categories[0]
        LOGGER.info("'%s' inferred as the first term of the year", first_term_of_year)
        assert isinstance(first_term_of_year, str)  # type guard
        return first_term_of_year  # type: ignore
    else:
        raise ValueError(
            f"'{s.name}' series is not an ordered categorical: {s.dtype=} ..."
            "so the first term of the academic year can't be inferred. "
            "Update the raw course data schema to properly order its categories!"
        )


def infer_num_terms_in_year(s: pd.Series) -> int:
    """
    Infer the number of terms in the (academic) year by the number of its categorical values.

    See Also:
        - :class:`schemas.base.TermField()`
    """
    if isinstance(s.dtype, pd.CategoricalDtype):
        num_terms_in_year = len(s.cat.categories)
        LOGGER.info("%s inferred as the number of term in the year", num_terms_in_year)
        return num_terms_in_year
    else:
        raise ValueError(
            f"'{s.name}' series is not a categorical: {s.dtype=} ..."
            "so the number of term in the academic year can't be inferred. "
            "Update the raw course data schema to properly set its categories!"
        )



import csv
import random
import string
from datetime import datetime, timedelta


# TODO: duplicated in synthetic_dataset.py
def generate_random_string(n):
    """Generate random alphanumeric string of length n"""
    return "".join(random.choices(string.ascii_uppercase + string.digits, k=n))


# TODO: duplicated in synthetic_dataset.py
def generate_random_date(start, end):
    """Generate a random date between start and end"""
    total_days = (end - start).days
    random_days = random.randint(0, total_days)
    return start + timedelta(days=random_days)


# TODO: duplicated in synthetic_dataset.py
def generate_sequential_dates(start_year, end_year, latest_possible_date, num_records):
    """Generate a sequence of random dates in ascending order, starting within a given range"""
    start_date = generate_random_date(
        datetime(start_year, 1, 1), datetime(end_year, 12, 31)
    )
    dates = [start_date]
    for _ in range(1, num_records):
        next_date = dates[-1] + timedelta(
            days=random.randint(30, 365)
        )  # Add 30 to 365 days to the last date
        if next_date > latest_possible_date:
            break
        dates.append(next_date)
    return dates


# TODO: duplicated in synthetic_data.py
def generate_random_decimal(min_value, max_value, n):
    """Generate a random value between min_value and max_value with n decimal places"""
    return round(random.uniform(min_value, max_value), n)


# List of possible College IDs and Program Titles
college_ids = [2, 3, 4, 8, 9, 12, 13, 11]
program_titles = [
    "POLICE SCIENCE",
    "BUSINESS ADMINISTRATION",
    "SECURITY MANAGEMENT",
    "PUBLIC ADMINISTRATION",
    "LAW AND SOCIETY",
    "WRITING AND LITERATURE",
    "LIBERAL ARTS AND SCIENCE",
    "RADIOLOGIC TECHNOLOGY",
    "LIB ARTS: SOCIAL SCIENCES & HUMANITIES",
    "FORENSIC PSYCHOLOGY",
    "COMPUTER OPERATIONS",
]

# Generate unique IDs
unique_ids = [
    generate_random_string(8) for _ in range(67)
]  # Approximately 67 unique IDs

# Open the CSV file for writing
with open("0424_23_synthetic_data_5.csv", mode="w", newline="") as file:
    writer = csv.writer(file)
    writer.writerow(
        [
            "UNIQUE_ID",
            "Term",
            "College ID",
            "Year Graduated",
            "Semester Graduated Desc",
            "Full Part Type Desc",
            "Credits+Hours Semester All Courses Show SUM",
            "Credits+Hours Semester Completed Perf SUM",
            "Credits+Hours Semester Passed Perf SUM",
            "Credits Semester Included in GPA Perf SUM",
            "IR Total FA Awd Amt SUM",
            "Ir Total Grants Nb Awd Amt SUM",
            "Ir Student Loans Nb Awd Amt SUM",
            "Ir Fed Work Study Nb Awd Amt SUM",
            "PELL_AMT_AWD",
            "Pell Flag",
            "TAP Flag",
            "TAP_AMT_AWD",
            "GPA Semester Perf",
            "GPA Cumulative Perf",
            "GPA Cumulative Show",
            "Major 1 NYSED Program Title",
            "Degree Earned Level Desc",
            "Value Points Semester Perf SUM",
        ]
    )

    for unique_id in unique_ids:
        num_records = random.randint(10, 15)  # Between 10 and 15 records per unique ID
        terms = generate_sequential_dates(
            2010, 2018, datetime(2020, 12, 31), num_records
        )

        for term in terms:
            term_str = term.strftime("%Y-%m-%d")
            college_id = random.choice(college_ids)
            year_graduated = ""
            semester_graduated_desc = ""
            full_part_type_desc = random.choice(["PART-TIME", "FULL-TIME"])
            credits_hours_semester_all_courses_show_sum = generate_random_decimal(
                6, 13, 1
            )
            credits_hours_semester_completed_perf_sum = generate_random_decimal(
                6, 13, 1
            )
            credits_hours_semester_passed_perf_sum = generate_random_decimal(3, 12, 1)
            credits_semester_included_in_gpa_perf_sum = generate_random_decimal(
                6, 12, 1
            )
            ir_total_fa_awd_amt_sum = generate_random_decimal(2800, 8200, 2)
            ir_total_grants_nb_awd_amt_sum = generate_random_decimal(950, 7777, 2)
            ir_student_loans_nb_awd_amt_sum = generate_random_decimal(3850, 5500, 2)
            ir_fed_work_study_nb_awd_amt_sum = 0
            pell_amt_awd = generate_random_decimal(650, 2600, 2)
            pell_flag = random.choice(["Yes", "No"])
            tap_flag = random.choice(["Yes", "No"])
            tap_amt_awd = generate_random_decimal(0, 1300, 2)
            gpa_semester_perf = generate_random_decimal(2.4, 3.1, 2)
            gpa_cumulative_perf = generate_random_decimal(2.4, 3, 2)
            gpa_cumulative_show = generate_random_decimal(2.4, 3.1, 2)
            major_1_nysed_program_title = random.choice(program_titles)
            degree_earned_level_desc = ""
            value_points_semester_perf_sum = generate_random_decimal(9, 38, 1)

            writer.writerow(
                [
                    unique_id,
                    term_str,
                    college_id,
                    year_graduated,
                    semester_graduated_desc,
                    full_part_type_desc,
                    credits_hours_semester_all_courses_show_sum,
                    credits_hours_semester_completed_perf_sum,
                    credits_hours_semester_passed_perf_sum,
                    credits_semester_included_in_gpa_perf_sum,
                    ir_total_fa_awd_amt_sum,
                    ir_total_grants_nb_awd_amt_sum,
                    ir_student_loans_nb_awd_amt_sum,
                    ir_fed_work_study_nb_awd_amt_sum,
                    pell_amt_awd,
                    pell_flag,
                    tap_flag,
                    tap_amt_awd,
                    gpa_semester_perf,
                    gpa_cumulative_perf,
                    gpa_cumulative_show,
                    major_1_nysed_program_title,
                    degree_earned_level_desc,
                    value_points_semester_perf_sum,
                ]
            )



import logging
import time

import pandas as pd
import pyspark.sql

from .. import utils

LOGGER = logging.getLogger(__name__)


def to_delta_table(
    df: pd.DataFrame, table_path: str, spark_session: pyspark.sql.SparkSession
) -> None:
    """
    Write pandas DataFrame to Databricks Unity Catalog.

    Args:
        df
        table_path: Path in Unity Catalog to which ``df`` will be written,
            including the full three-level namespace: ``catalog.schema.table`` .
        spark_session: Entry point to using spark dataframes and the databricks integration.

    See Also:
        - https://docs.databricks.com/en/delta/drop-table.html#when-to-replace-a-table
        - https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/spark_session.html
    """
    start_time = time.time()
    LOGGER.info("saving data to %s ...", table_path)
    df_spark = spark_session.createDataFrame(
        df.rename(columns=utils.misc.convert_to_snake_case)
    )
    (
        pyspark.sql.DataFrameWriterV2(df_spark, table_path)
        .options(format="delta")
        # this *should* do what databricks recomends -- and retains table history!
        .createOrReplace()
    )
    run_time = time.time() - start_time

    table_rows = spark_session.sql(f"SELECT COUNT(*) FROM {table_path}").collect()
    if table_rows[0][0] != len(df):
        raise IOError(
            f"{table_rows[0][0]} written to delta table, "
            f"but {len(df)} rows in original dataframe"
        )

    history = spark_session.sql(f"DESCRIBE history {table_path} LIMIT 1").collect()
    verno = int(history[0][0])
    LOGGER.info("data saved to %s (v%s) in %s seconds", table_path, verno, run_time)



import pytest
import pandas as pd
from unittest.mock import MagicMock, patch
from student_success_tool.reporting.sections.registry import SectionRegistry
from student_success_tool.reporting.sections.evaluation_sections import (
    register_evaluation_sections,
)
from student_success_tool.reporting.utils.formatting import Formatting


@pytest.fixture
def mock_card(tmp_path):
    card = MagicMock()
    card.run_id = "123"
    card.assets_folder = str(tmp_path)
    card.format = Formatting()
    return card


@patch(
    "student_success_tool.reporting.sections.evaluation_sections.utils.list_paths_in_directory"
)
@patch(
    "student_success_tool.reporting.sections.evaluation_sections.utils.download_artifact"
)
def test_register_evaluation_sections_success(
    mock_download, mock_list_paths, mock_card, tmp_path
):
    # Setup mock CSV file
    csv_path = tmp_path / "test_file.csv"
    df = pd.DataFrame({"Metric": ["Accuracy", "Recall"], "Value": [0.9, 0.85]})
    df.to_csv(csv_path, index=False)

    mock_list_paths.return_value = [
        "group_metrics/bias_test_gender_metrics.csv",
        "group_metrics/perf_test_gender_metrics.csv",
    ]
    mock_download.return_value = str(csv_path)

    registry = SectionRegistry()
    register_evaluation_sections(mock_card, registry)
    rendered = registry.render_all()
    print(rendered["evaluation_by_group_section"])
    assert (
        "Evaluation Metrics by Student Group" in rendered["evaluation_by_group_section"]
    )
    assert "Gender Bias Metrics" in rendered["evaluation_by_group_section"]
    assert "| Accuracy | 0.9 |" in rendered["evaluation_by_group_section"]


@patch(
    "student_success_tool.reporting.sections.evaluation_sections.utils.list_paths_in_directory"
)
@patch(
    "student_success_tool.reporting.sections.evaluation_sections.utils.download_artifact"
)
def test_register_evaluation_sections_failure(
    mock_download, mock_list_paths, mock_card
):
    mock_list_paths.return_value = [
        "group_metrics/bias_test_race_metrics.csv",
        "group_metrics/perf_test_race_metrics.csv",
    ]
    mock_download.side_effect = Exception("Download error")

    registry = SectionRegistry()
    register_evaluation_sections(mock_card, registry)
    rendered = registry.render_all()

    assert "Race Performance Metrics" in rendered["evaluation_by_group_section"]
    assert "Could not load data." in rendered["evaluation_by_group_section"]
